import os
import threading
import time
import uuid

import requests
import sqlalchemy as sa

# Public API
from dbos import DBOS, ConfigFile, Queue, SetWorkflowID, _workflow_commands
from dbos._schemas.system_database import SystemSchema
from dbos._sys_db import SystemDatabase, WorkflowStatusString
from dbos._utils import GlobalParams


def test_admin_endpoints(dbos: DBOS) -> None:

    # Test GET /dbos-healthz
    response = requests.get("http://localhost:3001/dbos-healthz", timeout=5)
    assert response.status_code == 200
    assert response.text == "healthy"

    # Test POST /dbos-workflow-recovery
    data = ["executor1", "executor2"]
    response = requests.post(
        "http://localhost:3001/dbos-workflow-recovery", json=data, timeout=5
    )
    assert response.status_code == 200
    assert response.json() == []

    # Test GET /dbos-workflow-queues-metadata
    Queue("q1")
    Queue("q2", concurrency=1)
    Queue("q3", concurrency=1, worker_concurrency=1)
    Queue("q4", concurrency=1, worker_concurrency=1, limiter={"limit": 0, "period": 0})
    response = requests.get(
        "http://localhost:3001/dbos-workflow-queues-metadata", timeout=5
    )
    assert response.status_code == 200
    assert response.json() == [
        {"name": "q1"},
        {"name": "q2", "concurrency": 1},
        {"name": "q3", "concurrency": 1, "workerConcurrency": 1},
        {
            "name": "q4",
            "concurrency": 1,
            "workerConcurrency": 1,
            "rateLimit": {"limit": 0, "period": 0},
        },
    ]

    # Test GET not found
    response = requests.get("http://localhost:3001/stuff", timeout=5)
    assert response.status_code == 404

    # Test POST not found
    response = requests.post("http://localhost:3001/stuff", timeout=5)
    assert response.status_code == 404

    response = requests.get("http://localhost:3001/deactivate", timeout=5)
    assert response.status_code == 200

    for event in dbos.stop_events:
        assert event.is_set(), "Event is not set!"


def test_admin_recovery(config: ConfigFile) -> None:
    os.environ["DBOS__VMID"] = "testexecutor"
    os.environ["DBOS__APPVERSION"] = "testversion"
    os.environ["DBOS__APPID"] = "testappid"
    DBOS.destroy(destroy_registry=True)
    dbos = DBOS(config=config)
    DBOS.launch()

    step_counter: int = 0
    wf_counter: int = 0

    @DBOS.workflow()
    def test_workflow(var: str, var2: str) -> str:
        DBOS.logger.info("WFID: " + DBOS.workflow_id)
        nonlocal wf_counter
        wf_counter += 1
        res = test_step(var2)
        return res + var

    @DBOS.step()
    def test_step(var2: str) -> str:
        nonlocal step_counter
        step_counter += 1
        return var2 + "1"

    wfuuid = str(uuid.uuid4())
    DBOS.logger.info("Initiating: " + wfuuid)
    with SetWorkflowID(wfuuid):
        assert test_workflow("bob", "bob") == "bob1bob"

    dbos._sys_db.wait_for_buffer_flush()

    # Manually update the database to pretend the workflow comes from another executor and is pending
    with dbos._sys_db.engine.begin() as c:
        query = (
            sa.update(SystemSchema.workflow_status)
            .values(
                status=WorkflowStatusString.PENDING.value, executor_id="other-executor"
            )
            .where(SystemSchema.workflow_status.c.workflow_uuid == wfuuid)
        )
        c.execute(query)

    status = dbos.get_workflow_status(wfuuid)
    assert (
        status is not None and status.status == "PENDING"
    ), "Workflow status not updated"

    # Test POST /dbos-workflow-recovery
    data = ["other-executor"]
    response = requests.post(
        "http://localhost:3001/dbos-workflow-recovery", json=data, timeout=5
    )
    assert response.status_code == 200
    assert response.json() == [wfuuid]

    # Wait until the workflow is recovered
    max_retries = 10
    succeeded = False
    for attempt in range(max_retries):
        status = dbos.get_workflow_status(wfuuid)
        if status is not None and status.status == "SUCCESS":
            assert status.status == "SUCCESS"
            assert status.executor_id == "testexecutor"
            succeeded = True
            break
        else:
            time.sleep(1)
            print(f"Attempt {attempt + 1} failed. Retrying in 1 second...")
    assert succeeded, "Workflow did not recover"


def test_admin_diff_port(cleanup_test_databases: None) -> None:
    # Initialize singleton
    DBOS.destroy()  # In case of other tests leaving it

    config_string = """name: test-app
language: python
database:
  hostname: localhost
  port: 5432
  username: postgres
  password: ${PGPASSWORD}
  app_db_name: dbostestpy
runtimeConfig:
  start:
    - python3 main.py
  admin_port: 8001
"""
    # Write the config to a text file for the moment
    with open("dbos-config.yaml", "w") as file:
        file.write(config_string)

    try:
        # Initialize DBOS
        DBOS()
        DBOS.launch()

        # Test GET /dbos-healthz
        response = requests.get("http://localhost:8001/dbos-healthz", timeout=5)
        assert response.status_code == 200
        assert response.text == "healthy"
    finally:
        # Clean up after the test
        DBOS.destroy()
        os.remove("dbos-config.yaml")


def test_admin_workflow_resume(dbos: DBOS, sys_db: SystemDatabase) -> None:
    counter: int = 0
    event = threading.Event()

    @DBOS.workflow()
    def simple_workflow() -> None:
        event.set()
        nonlocal counter
        counter += 1

    # Run the workflow and flush its results
    simple_workflow()
    assert counter == 1
    dbos._sys_db.wait_for_buffer_flush()

    # Verify the workflow has succeeded
    output = _workflow_commands.list_workflows(sys_db)
    assert len(output) == 1, f"Expected list length to be 1, but got {len(output)}"
    assert output[0] != None, "Expected output to be not None"
    wfUuid = output[0].workflow_id
    info = _workflow_commands.get_workflow(sys_db, wfUuid, True)
    assert info is not None, "Expected output to be not None"
    assert info.status == "SUCCESS", f"Expected status to be SUCCESS"

    # Cancel the workflow. Verify it was cancelled
    response = requests.post(
        f"http://localhost:3001/workflows/{wfUuid}/cancel", json=[], timeout=5
    )
    assert response.status_code == 204
    info = _workflow_commands.get_workflow(sys_db, wfUuid, True)
    assert info is not None
    assert info.status == "CANCELLED", f"Expected status to be CANCELLED"

    # Manually update the database to pretend the workflow comes from another executor and is pending
    with dbos._sys_db.engine.begin() as c:
        query = (
            sa.update(SystemSchema.workflow_status)
            .values(
                status=WorkflowStatusString.PENDING.value, executor_id="other-executor"
            )
            .where(SystemSchema.workflow_status.c.workflow_uuid == wfUuid)
        )
        c.execute(query)

    # Resume the workflow. Verify that it succeeds again.
    event.clear()
    response = requests.post(
        f"http://localhost:3001/workflows/{wfUuid}/resume", json=[], timeout=5
    )
    assert response.status_code == 204
    assert event.wait(timeout=5)
    dbos._sys_db.wait_for_buffer_flush()
    assert counter == 2
    info = _workflow_commands.get_workflow(sys_db, wfUuid, True)
    assert info is not None
    assert info.status == "SUCCESS", f"Expected status to be SUCCESS"
    assert info.executor_id == GlobalParams.executor_id

    # Resume the workflow. Verify it does not run and status remains SUCCESS
    response = requests.post(
        f"http://localhost:3001/workflows/{wfUuid}/resume", json=[], timeout=5
    )
    assert response.status_code == 204
    dbos._sys_db.wait_for_buffer_flush()
    info = _workflow_commands.get_workflow(sys_db, wfUuid, True)
    assert info is not None
    assert info.status == "SUCCESS", f"Expected status to be SUCCESS"
    assert counter == 2


def test_admin_workflow_restart(dbos: DBOS, sys_db: SystemDatabase) -> None:

    @DBOS.workflow()
    def simple_workflow() -> None:
        print("Executed Simple workflow")
        return

    # run the workflow
    simple_workflow()
    time.sleep(1)

    # get the workflow list
    output = _workflow_commands.list_workflows(sys_db)
    assert len(output) == 1, f"Expected list length to be 1, but got {len(output)}"

    assert output[0] != None, "Expected output to be not None"

    wfUuid = output[0].workflow_id

    info = _workflow_commands.get_workflow(sys_db, wfUuid, True)
    assert info is not None, "Expected output to be not None"

    assert info.status == "SUCCESS", f"Expected status to be SUCCESS"

    response = requests.post(
        f"http://localhost:3001/workflows/{wfUuid}/cancel", json=[], timeout=5
    )
    assert response.status_code == 204

    info = _workflow_commands.get_workflow(sys_db, wfUuid, True)
    if info is not None:
        assert info.status == "CANCELLED", f"Expected status to be CANCELLED"
    else:
        assert False, "Expected info to be not None"

    response = requests.post(
        f"http://localhost:3001/workflows/{wfUuid}/restart", json=[], timeout=5
    )
    assert response.status_code == 204

    time.sleep(1)

    info = _workflow_commands.get_workflow(sys_db, wfUuid, True)
    if info is not None:
        assert info.status == "CANCELLED", f"Expected status to be CANCELLED"
    else:
        assert False, "Expected info to be not None"

    output = _workflow_commands.list_workflows(sys_db)
    assert len(output) == 2, f"Expected list length to be 2, but got {len(output)}"

    if output[0].workflow_id == wfUuid:
        new_wfUuid = output[1].workflow_id
    else:
        new_wfUuid = output[0].workflow_id

    info = _workflow_commands.get_workflow(sys_db, new_wfUuid, True)
    if info is not None:
        assert info.status == "SUCCESS", f"Expected status to be SUCCESS"
    else:
        assert False, "Expected info to be not None"



from typing import Optional, TypedDict

import sqlalchemy as sa
import sqlalchemy.dialects.postgresql as pg
from sqlalchemy.exc import DBAPIError
from sqlalchemy.orm import Session, sessionmaker

from ._dbos_config import ConfigFile
from ._error import DBOSWorkflowConflictIDError
from ._schemas.application_database import ApplicationSchema


class TransactionResultInternal(TypedDict):
    workflow_uuid: str
    function_id: int
    output: Optional[str]  # JSON (jsonpickle)
    error: Optional[str]  # JSON (jsonpickle)
    txn_id: Optional[str]
    txn_snapshot: str
    executor_id: Optional[str]


class RecordedResult(TypedDict):
    output: Optional[str]  # JSON (jsonpickle)
    error: Optional[str]  # JSON (jsonpickle)


class ApplicationDatabase:

    def __init__(self, config: ConfigFile, *, debug_mode: bool = False):
        self.config = config

        app_db_name = config["database"]["app_db_name"]

        # If the application database does not already exist, create it
        if not debug_mode:
            postgres_db_url = sa.URL.create(
                "postgresql+psycopg",
                username=config["database"]["username"],
                password=config["database"]["password"],
                host=config["database"]["hostname"],
                port=config["database"]["port"],
                database="postgres",
            )
            postgres_db_engine = sa.create_engine(postgres_db_url)
            with postgres_db_engine.connect() as conn:
                conn.execution_options(isolation_level="AUTOCOMMIT")
                if not conn.execute(
                    sa.text("SELECT 1 FROM pg_database WHERE datname=:db_name"),
                    parameters={"db_name": app_db_name},
                ).scalar():
                    conn.execute(sa.text(f"CREATE DATABASE {app_db_name}"))
            postgres_db_engine.dispose()

        # Create a connection pool for the application database
        app_db_url = sa.URL.create(
            "postgresql+psycopg",
            username=config["database"]["username"],
            password=config["database"]["password"],
            host=config["database"]["hostname"],
            port=config["database"]["port"],
            database=app_db_name,
        )
        self.engine = sa.create_engine(
            app_db_url, pool_size=20, max_overflow=5, pool_timeout=30
        )
        self.sessionmaker = sessionmaker(bind=self.engine)
        self.debug_mode = debug_mode

        # Create the dbos schema and transaction_outputs table in the application database
        if not debug_mode:
            with self.engine.begin() as conn:
                schema_creation_query = sa.text(
                    f"CREATE SCHEMA IF NOT EXISTS {ApplicationSchema.schema}"
                )
                conn.execute(schema_creation_query)
            ApplicationSchema.metadata_obj.create_all(self.engine)

    def destroy(self) -> None:
        self.engine.dispose()

    @staticmethod
    def record_transaction_output(
        session: Session, output: TransactionResultInternal
    ) -> None:
        try:
            session.execute(
                pg.insert(ApplicationSchema.transaction_outputs).values(
                    workflow_uuid=output["workflow_uuid"],
                    function_id=output["function_id"],
                    output=output["output"],
                    error=None,
                    txn_id=sa.text("(select pg_current_xact_id_if_assigned()::text)"),
                    txn_snapshot=output["txn_snapshot"],
                    executor_id=(
                        output["executor_id"] if output["executor_id"] else None
                    ),
                )
            )
        except DBAPIError as dbapi_error:
            if dbapi_error.orig.sqlstate == "23505":  # type: ignore
                raise DBOSWorkflowConflictIDError(output["workflow_uuid"])
            raise

    def record_transaction_error(self, output: TransactionResultInternal) -> None:
        if self.debug_mode:
            raise Exception("called record_transaction_error in debug mode")
        try:
            with self.engine.begin() as conn:
                conn.execute(
                    pg.insert(ApplicationSchema.transaction_outputs).values(
                        workflow_uuid=output["workflow_uuid"],
                        function_id=output["function_id"],
                        output=None,
                        error=output["error"],
                        txn_id=sa.text(
                            "(select pg_current_xact_id_if_assigned()::text)"
                        ),
                        txn_snapshot=output["txn_snapshot"],
                        executor_id=(
                            output["executor_id"] if output["executor_id"] else None
                        ),
                    )
                )
        except DBAPIError as dbapi_error:
            if dbapi_error.orig.sqlstate == "23505":  # type: ignore
                raise DBOSWorkflowConflictIDError(output["workflow_uuid"])
            raise

    @staticmethod
    def check_transaction_execution(
        session: Session, workflow_uuid: str, function_id: int
    ) -> Optional[RecordedResult]:
        rows = session.execute(
            sa.select(
                ApplicationSchema.transaction_outputs.c.output,
                ApplicationSchema.transaction_outputs.c.error,
            ).where(
                ApplicationSchema.transaction_outputs.c.workflow_uuid == workflow_uuid,
                ApplicationSchema.transaction_outputs.c.function_id == function_id,
            )
        ).all()
        if len(rows) == 0:
            return None
        result: RecordedResult = {
            "output": rows[0][0],
            "error": rows[0][1],
        }
        return result



import threading
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Callable

from ._logger import dbos_logger
from ._queue import Queue

if TYPE_CHECKING:
    from ._dbos import DBOSRegistry

from ._context import SetWorkflowID
from ._croniter import croniter  # type: ignore

ScheduledWorkflow = Callable[[datetime, datetime], None]

scheduler_queue: Queue


def scheduler_loop(
    func: ScheduledWorkflow, cron: str, stop_event: threading.Event
) -> None:
    try:
        iter = croniter(cron, datetime.now(timezone.utc), second_at_beginning=True)
    except Exception as e:
        dbos_logger.error(
            f'Cannot run scheduled function {func.__name__}. Invalid crontab "{cron}"'
        )
    while not stop_event.is_set():
        nextExecTime = iter.get_next(datetime)
        sleepTime = nextExecTime - datetime.now(timezone.utc)
        if stop_event.wait(timeout=sleepTime.total_seconds()):
            return
        with SetWorkflowID(f"sched-{func.__qualname__}-{nextExecTime.isoformat()}"):
            try:
                scheduler_queue.enqueue(func, nextExecTime, datetime.now(timezone.utc))
            except Exception as e:
                dbos_logger.warning(f"Error scheduling workflow: ", e)


def scheduled(
    dbosreg: "DBOSRegistry", cron: str
) -> Callable[[ScheduledWorkflow], ScheduledWorkflow]:
    def decorator(func: ScheduledWorkflow) -> ScheduledWorkflow:
        try:
            croniter(cron, datetime.now(timezone.utc), second_at_beginning=True)
        except Exception as e:
            raise ValueError(
                f'Invalid crontab "{cron}" for scheduled function function {func.__name__}.'
            )

        global scheduler_queue
        scheduler_queue = Queue("_dbos_internal_queue")
        stop_event = threading.Event()
        dbosreg.register_poller(stop_event, scheduler_loop, func, cron, stop_event)
        return func

    return decorator



from sqlalchemy import Column, Integer, MetaData, String, Table

metadata = MetaData()

dbos_hello = Table(
    "dbos_hello",
    metadata,
    Column("greet_count", Integer, primary_key=True, autoincrement=True),
    Column("name", String, nullable=False),
)



import asyncio
import contextlib
import inspect
import time
from typing import Any, Callable, Coroutine, Optional, Protocol, TypeVar, Union, cast

T = TypeVar("T")
R = TypeVar("R")


class NoResult:
    _instance: Optional["NoResult"] = None
    __slots__ = ()

    def __new__(cls, *args: Any, **kwargs: Any) -> "NoResult":
        if not cls._instance:
            cls._instance = super(NoResult, cls).__new__(cls, *args, **kwargs)
        return cls._instance


# define Outcome protocol w/ common composition methods
class Outcome(Protocol[T]):

    def wrap(
        self, before: Callable[[], Callable[[Callable[[], T]], R]]
    ) -> "Outcome[R]": ...

    def then(self, next: Callable[[Callable[[], T]], R]) -> "Outcome[R]": ...

    def also(
        self, cm: contextlib.AbstractContextManager[Any, bool]
    ) -> "Outcome[T]": ...

    def retry(
        self,
        attempts: int,
        on_exception: Callable[[int, BaseException], float],
        exceeded_retries: Callable[[int], BaseException],
    ) -> "Outcome[T]": ...

    def intercept(
        self, interceptor: Callable[[], Union[NoResult, T]]
    ) -> "Outcome[T]": ...

    def __call__(self) -> Union[T, Coroutine[Any, Any, T]]: ...

    # Helper function to create an Immediate or Pending Result, depending on if func is a coroutine function or not
    @staticmethod
    def make(func: Callable[[], Union[T, Coroutine[Any, Any, T]]]) -> "Outcome[T]":
        return (
            Pending(cast(Callable[[], Coroutine[Any, Any, T]], func))
            if inspect.iscoroutinefunction(func)
            else Immediate(cast(Callable[[], T], func))
        )


# Immediate Outcome - for composing non-async functions
class Immediate(Outcome[T]):
    __slots__ = "_func"

    def __init__(self, func: Callable[[], T]):
        self._func = func

    def then(self, next: Callable[[Callable[[], T]], R]) -> "Immediate[R]":
        return Immediate(lambda: next(self._func))

    def wrap(
        self, before: Callable[[], Callable[[Callable[[], T]], R]]
    ) -> "Immediate[R]":
        return Immediate(lambda: before()(self._func))

    @staticmethod
    def _intercept(
        func: Callable[[], T], interceptor: Callable[[], Union[NoResult, T]]
    ) -> T:
        intercepted = interceptor()
        return intercepted if not isinstance(intercepted, NoResult) else func()

    def intercept(
        self, interceptor: Callable[[], Union[NoResult, T]]
    ) -> "Immediate[T]":
        return Immediate[T](lambda: Immediate._intercept(self._func, interceptor))

    @staticmethod
    def _also(func: Callable[[], T], cm: contextlib.AbstractContextManager[Any, bool]) -> T:  # type: ignore
        with cm:
            return func()

    def also(self, cm: contextlib.AbstractContextManager[Any, bool]) -> "Immediate[T]":
        return Immediate[T](lambda: Immediate._also(self._func, cm))

    @staticmethod
    def _retry(
        func: Callable[[], T],
        attempts: int,
        on_exception: Callable[[int, BaseException], float],
        exceeded_retries: Callable[[int], BaseException],
    ) -> T:
        for i in range(attempts):
            try:
                return func()
            except Exception as exp:
                wait_time = on_exception(i, exp)
                time.sleep(wait_time)

        raise exceeded_retries(attempts)

    def retry(
        self,
        attempts: int,
        on_exception: Callable[[int, BaseException], float],
        exceeded_retries: Callable[[int], BaseException],
    ) -> "Immediate[T]":
        assert attempts > 0
        return Immediate[T](
            lambda: Immediate._retry(
                self._func, attempts, on_exception, exceeded_retries
            )
        )

    def __call__(self) -> T:
        return self._func()


# Pending Outcome - for composing async functions
class Pending(Outcome[T]):
    __slots__ = "_func"

    def __init__(self, func: Callable[[], Coroutine[Any, Any, T]]):
        self._func = func

    # Helper method in order to raise an exception in a lambda
    @staticmethod
    def _raise(ex: BaseException) -> T:
        raise ex

    async def _wrap(
        func: Callable[[], Coroutine[Any, Any, T]],
        before: Callable[[], Callable[[Callable[[], T]], R]],
    ) -> R:
        after = await asyncio.to_thread(before)
        try:
            value = await func()
            return await asyncio.to_thread(after, lambda: value)
        except BaseException as exp:
            return await asyncio.to_thread(after, lambda: Pending._raise(exp))

    def wrap(
        self, before: Callable[[], Callable[[Callable[[], T]], R]]
    ) -> "Pending[R]":
        return Pending[R](lambda: Pending._wrap(self._func, before))

    def then(self, next: Callable[[Callable[[], T]], R]) -> "Pending[R]":
        return Pending[R](lambda: Pending._wrap(self._func, lambda: next))

    @staticmethod
    async def _also(  # type: ignore
        func: Callable[[], Coroutine[Any, Any, T]],
        cm: contextlib.AbstractContextManager[Any, bool],
    ) -> T:
        with cm:
            return await func()

    def also(self, cm: contextlib.AbstractContextManager[Any, bool]) -> "Pending[T]":
        return Pending[T](lambda: Pending._also(self._func, cm))

    @staticmethod
    async def _intercept(
        func: Callable[[], Coroutine[Any, Any, T]],
        interceptor: Callable[[], Union[NoResult, T]],
    ) -> T:
        intercepted = await asyncio.to_thread(interceptor)
        return intercepted if not isinstance(intercepted, NoResult) else await func()

    def intercept(self, interceptor: Callable[[], Union[NoResult, T]]) -> "Pending[T]":
        return Pending[T](lambda: Pending._intercept(self._func, interceptor))

    @staticmethod
    async def _retry(
        func: Callable[[], Coroutine[Any, Any, T]],
        attempts: int,
        on_exception: Callable[[int, BaseException], float],
        exceeded_retries: Callable[[int], BaseException],
    ) -> T:
        for i in range(attempts):
            try:
                return await func()
            except Exception as exp:
                wait_time = on_exception(i, exp)
                await asyncio.sleep(wait_time)

        raise exceeded_retries(attempts)

    def retry(
        self,
        attempts: int,
        on_exception: Callable[[int, BaseException], float],
        exceeded_retries: Callable[[int], BaseException],
    ) -> "Pending[T]":
        assert attempts > 0
        return Pending[T](
            lambda: Pending._retry(self._func, attempts, on_exception, exceeded_retries)
        )

    async def __call__(self) -> T:
        return await self._func()



import os
import threading
import time
import traceback
from typing import TYPE_CHECKING, Any, List

from dbos._utils import GlobalParams

from ._core import execute_workflow_by_id
from ._error import DBOSWorkflowFunctionNotFoundError
from ._sys_db import GetPendingWorkflowsOutput

if TYPE_CHECKING:
    from ._dbos import DBOS, WorkflowHandle


def _recover_workflow(
    dbos: "DBOS", workflow: GetPendingWorkflowsOutput
) -> "WorkflowHandle[Any]":
    if workflow.queue_name and workflow.queue_name != "_dbos_internal_queue":
        cleared = dbos._sys_db.clear_queue_assignment(workflow.workflow_uuid)
        if cleared:
            return dbos.retrieve_workflow(workflow.workflow_uuid)
    return execute_workflow_by_id(dbos, workflow.workflow_uuid)


def startup_recovery_thread(
    dbos: "DBOS", pending_workflows: List[GetPendingWorkflowsOutput]
) -> None:
    """Attempt to recover local pending workflows on startup using a background thread."""
    stop_event = threading.Event()
    dbos.stop_events.append(stop_event)
    while not stop_event.is_set() and len(pending_workflows) > 0:
        try:
            for pending_workflow in list(pending_workflows):
                _recover_workflow(dbos, pending_workflow)
                pending_workflows.remove(pending_workflow)
        except DBOSWorkflowFunctionNotFoundError:
            time.sleep(1)
        except Exception as e:
            dbos.logger.error(
                f"Exception encountered when recovering workflows: {traceback.format_exc()}"
            )
            raise e


def recover_pending_workflows(
    dbos: "DBOS", executor_ids: List[str] = ["local"]
) -> List["WorkflowHandle[Any]"]:
    """Attempt to recover pending workflows for a list of specific executors and return workflow handles for them."""
    workflow_handles: List["WorkflowHandle[Any]"] = []
    for executor_id in executor_ids:
        pending_workflows = dbos._sys_db.get_pending_workflows(
            executor_id, GlobalParams.app_version
        )
        for pending_workflow in pending_workflows:
            try:
                handle = _recover_workflow(dbos, pending_workflow)
                workflow_handles.append(handle)
            except Exception as e:
                dbos.logger.error(
                    f"Exception encountered when recovering workflows: {traceback.format_exc()}"
                )
                raise e
        dbos.logger.info(
            f"Recovering {len(pending_workflows)} workflows for executor {executor_id} from version {GlobalParams.app_version}"
        )
    return workflow_handles



import inspect
from dataclasses import dataclass
from enum import Enum
from types import FunctionType
from typing import Any, Callable, List, Literal, Optional, Tuple, Type, cast

DEFAULT_MAX_RECOVERY_ATTEMPTS = 50


def get_dbos_func_name(f: Any) -> str:
    if hasattr(f, "dbos_function_name"):
        return str(getattr(f, "dbos_function_name"))
    if hasattr(f, "__qualname__"):
        return str(getattr(f, "__qualname__"))
    return "<unknown>"


def set_dbos_func_name(f: Any, name: str) -> None:
    setattr(f, "dbos_function_name", name)


TempWorkflowType = Literal["transaction", "step", "send", None]


def get_temp_workflow_type(f: Any) -> TempWorkflowType:
    if hasattr(f, "dbos_temp_workflow_type"):
        return cast(TempWorkflowType, str(getattr(f, "dbos_temp_workflow_type")))
    return None


def set_temp_workflow_type(f: Any, name: TempWorkflowType) -> None:
    setattr(f, "dbos_temp_workflow_type", name)


@dataclass
class DBOSClassInfo:
    def_required_roles: Optional[List[str]] = None


class DBOSFuncType(Enum):
    Unknown = 0
    Bare = 1
    Static = 2
    Class = 3
    Instance = 4


@dataclass
class DBOSFuncInfo:
    class_info: Optional[DBOSClassInfo] = None
    func_type: DBOSFuncType = DBOSFuncType.Unknown
    required_roles: Optional[List[str]] = None
    max_recovery_attempts: int = DEFAULT_MAX_RECOVERY_ATTEMPTS


def get_or_create_class_info(cls: Type[Any]) -> DBOSClassInfo:
    if hasattr(cls, "dbos_class_decorator_info"):
        ci: DBOSClassInfo = getattr(cls, "dbos_class_decorator_info")
        return ci
    ci = DBOSClassInfo()
    setattr(cls, "dbos_class_decorator_info", ci)

    # Tell all DBOS functions about this
    for name, attribute in cls.__dict__.items():
        # Check if the attribute is a function or method
        if isinstance(attribute, (FunctionType, staticmethod, classmethod)):
            dbft = DBOSFuncType.Unknown
            if isinstance(attribute, staticmethod):
                attribute = attribute.__func__
                dbft = DBOSFuncType.Static
            elif isinstance(attribute, classmethod):
                attribute = attribute.__func__
                dbft = DBOSFuncType.Class
            elif isinstance(attribute, FunctionType):
                dbft = DBOSFuncType.Instance

            # Walk down the __wrapped__ chain
            wrapped = attribute
            while True:
                # Annotate based on the type
                if hasattr(wrapped, "dbos_func_decorator_info"):
                    fi: DBOSFuncInfo = getattr(wrapped, "dbos_func_decorator_info")
                    fi.class_info = ci
                    fi.func_type = dbft

                if not hasattr(wrapped, "__wrapped__"):
                    break
                wrapped = wrapped.__wrapped__

    return ci


def get_func_info(func: Callable[..., Any]) -> Optional[DBOSFuncInfo]:
    while True:
        if hasattr(func, "dbos_func_decorator_info"):
            fi: DBOSFuncInfo = getattr(func, "dbos_func_decorator_info")
            return fi
        if not hasattr(func, "__wrapped__"):
            break
        func = func.__wrapped__
    return None


def get_or_create_func_info(func: Callable[..., Any]) -> DBOSFuncInfo:
    fi = get_func_info(func)
    if fi is not None:
        return fi

    fi = DBOSFuncInfo()
    setattr(func, "dbos_func_decorator_info", fi)
    return fi


def set_func_info(func: Callable[..., Any], fi: DBOSFuncInfo) -> None:
    setattr(func, "dbos_func_decorator_info", fi)


def get_class_info(cls: Type[Any]) -> Optional[DBOSClassInfo]:
    if hasattr(cls, "dbos_class_decorator_info"):
        ci: DBOSClassInfo = getattr(cls, "dbos_class_decorator_info")
        return ci
    return None


def get_class_info_for_func(fi: Optional[DBOSFuncInfo]) -> Optional[DBOSClassInfo]:
    if fi and fi.class_info:
        return fi.class_info

    # Bare function or function on something else
    return None


def get_config_name(
    fi: Optional[DBOSFuncInfo], func: Callable[..., Any], args: Tuple[Any, ...]
) -> Optional[str]:
    if fi and fi.func_type != DBOSFuncType.Unknown and len(args) > 0:
        if fi.func_type == DBOSFuncType.Instance:
            first_arg = args[0]
            if hasattr(first_arg, "config_name"):
                iname: str = getattr(first_arg, "config_name")
                return str(iname)
            else:
                raise Exception(
                    "Function target appears to be a class instance, but does not have `config_name` set"
                )
        return None

    # Check for improperly-registered functions
    if len(args) > 0:
        first_arg = args[0]
        if isinstance(first_arg, type):
            raise Exception(
                "Function target appears to be a class, but is not properly registered"
            )
        else:
            # Check if the function signature has "self" as the first parameter name
            #   This is not 100% reliable but it is better than nothing for detecting footguns
            sig = inspect.signature(func)
            parameters = list(sig.parameters.values())
            if parameters and parameters[0].name == "self":
                raise Exception(
                    "Function target appears to be a class instance, but is not properly registered"
                )

    # Bare function or function on something else
    return None


def get_dbos_class_name(
    fi: Optional[DBOSFuncInfo], func: Callable[..., Any], args: Tuple[Any, ...]
) -> Optional[str]:
    if fi and fi.func_type != DBOSFuncType.Unknown and len(args) > 0:
        if fi.func_type == DBOSFuncType.Instance:
            first_arg = args[0]
            return str(first_arg.__class__.__name__)
        if fi.func_type == DBOSFuncType.Class:
            first_arg = args[0]
            return str(first_arg.__name__)
        return None

    # Check for improperly-registered functions
    if len(args) > 0:
        first_arg = args[0]
        if isinstance(first_arg, type):
            raise Exception(
                "Function target appears to be a class, but is not properly registered"
            )
        else:
            # Check if the function signature has "self" as the first parameter name
            #   This is not 100% reliable but it is better than nothing for detecting footguns
            sig = inspect.signature(func)
            parameters = list(sig.parameters.values())
            if parameters and parameters[0].name == "self":
                raise Exception(
                    "Function target appears to be a class instance, but is not properly registered"
                )

    # Bare function or function on something else
    return None



import sqlalchemy as sa

# Public API
from dbos import DBOS, DBOSConfiguredInstance


@DBOS.dbos_class()
class DBOSTestClass(DBOSConfiguredInstance):
    txn_counter_c = 0
    wf_counter_c = 0
    step_counter_c = 0

    def __init__(self) -> None:
        super().__init__("myconfig")
        self.txn_counter: int = 0
        self.wf_counter: int = 0
        self.step_counter: int = 0

    @classmethod
    @DBOS.workflow()
    def test_workflow_cls(cls, var: str, var2: str) -> str:
        cls.wf_counter_c += 1
        res = DBOSTestClass.test_transaction_cls(var2)
        res2 = DBOSTestClass.test_step_cls(var)
        return res + res2

    @classmethod
    @DBOS.transaction()
    def test_transaction_cls(cls, var2: str) -> str:
        rows = DBOS.sql_session.execute(sa.text("SELECT 1")).fetchall()
        cls.txn_counter_c += 1
        return var2 + str(rows[0][0])

    @classmethod
    @DBOS.step()
    def test_step_cls(cls, var: str) -> str:
        cls.step_counter_c += 1
        return var

    @DBOS.workflow()
    def test_workflow(self, var: str, var2: str) -> str:
        self.wf_counter += 1
        res = self.test_transaction(var2)
        res2 = self.test_step(var)
        return res + res2

    @DBOS.transaction()
    def test_transaction(self, var2: str) -> str:
        rows = DBOS.sql_session.execute(sa.text("SELECT 1")).fetchall()
        self.txn_counter += 1
        return var2 + str(rows[0][0])

    @DBOS.step()
    def test_step(self, var: str) -> str:
        self.step_counter += 1
        return var

    @DBOS.workflow()
    @DBOS.required_roles(["admin"])
    def test_func_admin(self, var: str) -> str:
        assert DBOS.assumed_role == "admin"
        return self.config_name + ":" + var


@DBOS.default_required_roles(["user"])
class DBOSTestRoles(DBOSConfiguredInstance):
    @staticmethod
    @DBOS.workflow()
    def greetfunc(name: str) -> str:
        return f"Hello {name}"


@DBOS.dbos_class()
class DBOSSendRecv:
    send_counter: int = 0
    recv_counter: int = 0

    @staticmethod
    @DBOS.workflow()
    def test_send_workflow(dest_uuid: str, topic: str) -> str:
        DBOS.send(dest_uuid, "test1")
        DBOS.send(dest_uuid, "test2", topic=topic)
        DBOS.send(dest_uuid, "test3")
        DBOSSendRecv.send_counter += 1
        return dest_uuid

    @staticmethod
    @DBOS.workflow()
    def test_recv_workflow(topic: str) -> str:
        msg1 = DBOS.recv(topic, timeout_seconds=10)
        msg2 = DBOS.recv(timeout_seconds=10)
        msg3 = DBOS.recv(timeout_seconds=10)
        DBOSSendRecv.recv_counter += 1
        return "-".join([str(msg1), str(msg2), str(msg3)])



import json
import os
import re
import time
from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional, Union

import jwt
import requests
import typer
from rich import print

from dbos._error import DBOSInitializationError

from .._logger import dbos_logger
from .authentication import authenticate


# Must be the same as in TS
@dataclass
class DBOSCloudCredentials:
    token: str
    userName: str
    organization: str
    refreshToken: Optional[str] = None


@dataclass
class UserProfile:
    Name: str
    Organization: str

    def __init__(self, **kwargs: Any) -> None:
        self.Name = kwargs.get("Name", "")
        self.Organization = kwargs.get("Organization", "")


class AppLanguages(Enum):
    Node = "node"
    Python = "python"


dbos_config_file_path = "dbos-config.yaml"
DBOS_CLOUD_HOST = os.getenv("DBOS_DOMAIN", "cloud.dbos.dev")
dbos_env_path = ".dbos"


def is_token_expired(token: str) -> bool:
    try:
        decoded = jwt.decode(token, options={"verify_signature": False})
        exp: int = decoded.get("exp")
        if not exp:
            return False
        return time.time() >= exp
    except Exception:
        return True


def credentials_exist() -> bool:
    return os.path.exists(os.path.join(dbos_env_path, "credentials"))


def delete_credentials() -> None:
    credentials_path = os.path.join(dbos_env_path, "credentials")
    if os.path.exists(credentials_path):
        os.unlink(credentials_path)


def write_credentials(credentials: DBOSCloudCredentials) -> None:
    os.makedirs(dbos_env_path, exist_ok=True)
    with open(os.path.join(dbos_env_path, "credentials"), "w", encoding="utf-8") as f:
        json.dump(credentials.__dict__, f)


def check_read_file(path: str, encoding: str = "utf-8") -> Union[str, bytes]:
    # Check if file exists and is a file
    if not os.path.exists(path):
        raise FileNotFoundError(f"File {path} does not exist")
    if not os.path.isfile(path):
        raise IsADirectoryError(f"Path {path} is not a file")

    # Read file content
    with open(path, encoding=encoding) as f:
        return f.read()


@dataclass
class CloudAPIErrorResponse:
    message: str
    status_code: int
    request_id: str
    detailed_error: Optional[str] = None


def is_cloud_api_error_response(obj: Any) -> bool:
    return (
        isinstance(obj, dict)
        and "message" in obj
        and isinstance(obj["message"], str)
        and "statusCode" in obj
        and isinstance(obj["statusCode"], int)
        and "requestID" in obj
        and isinstance(obj["requestID"], str)
    )


def handle_api_errors(label: str, e: requests.exceptions.RequestException) -> None:
    if hasattr(e, "response") and e.response is not None:
        resp = e.response.json()
        if is_cloud_api_error_response(resp):
            message = f"[{resp['requestID']}] {label}: {resp['message']}."
            dbos_logger.error(message)
            raise DBOSInitializationError(message)


def is_valid_username(value: str) -> Union[bool, str]:
    if len(value) < 3 or len(value) > 30:
        return "Username must be 3~30 characters long"
    if not re.match("^[a-z0-9_]+$", value):
        return "Username must contain only lowercase letters, numbers, and underscores."
    return True


def check_user_profile(credentials: DBOSCloudCredentials) -> bool:
    bearer_token = f"Bearer {credentials.token}"
    try:
        response = requests.get(
            f"https://{DBOS_CLOUD_HOST}/v1alpha1/user/profile",
            headers={
                "Content-Type": "application/json",
                "Authorization": bearer_token,
            },
        )
        response.raise_for_status()
        profile = UserProfile(**response.json())
        credentials.userName = profile.Name
        credentials.organization = profile.Organization
        return True
    except requests.exceptions.RequestException as e:
        error_label = "Failed to login"
        if hasattr(e, "response") and e.response is not None:
            resp = e.response.json()
            if is_cloud_api_error_response(resp):
                if "user not found in DBOS Cloud" not in resp["message"]:
                    handle_api_errors(error_label, e)
                    exit(1)
        else:
            dbos_logger.error(f"{error_label}: {str(e)}")
            exit(1)
    return False


def register_user(credentials: DBOSCloudCredentials) -> None:
    print("Please register for DBOS Cloud")

    user_name = None
    while not user_name:
        user_name = typer.prompt("Choose your username")
        validation_result = is_valid_username(user_name)
        if validation_result is not True:
            print(f"[red]Invalid username: {validation_result}[/red]")
            user_name = None
            continue

    bearer_token = f"Bearer {credentials.token}"
    try:
        # Register user
        response = requests.put(
            f"https://{DBOS_CLOUD_HOST}/v1alpha1/user",
            json={
                "name": user_name,
            },
            headers={
                "Content-Type": "application/json",
                "Authorization": bearer_token,
            },
        )
        response.raise_for_status()

        # Get user profile
        response = requests.get(
            f"https://{DBOS_CLOUD_HOST}/v1alpha1/user/profile",
            headers={
                "Content-Type": "application/json",
                "Authorization": bearer_token,
            },
        )
        response.raise_for_status()
        profile = UserProfile(**response.json())
        credentials.userName = profile.Name
        credentials.organization = profile.Organization
        print(f"[green]Successfully registered as {credentials.userName}[/green]")

    except requests.exceptions.RequestException as e:
        error_label = f"Failed to register user {user_name}"
        if hasattr(e, "response") and e.response is not None:
            handle_api_errors(error_label, e)
        else:
            dbos_logger.error(f"{error_label}: {str(e)}")
        exit(1)


def check_credentials() -> DBOSCloudCredentials:
    empty_credentials = DBOSCloudCredentials(token="", userName="", organization="")

    if not credentials_exist():
        return empty_credentials

    try:
        with open(os.path.join(dbos_env_path, "credentials"), "r") as f:
            cred_data = json.load(f)
            credentials = DBOSCloudCredentials(**cred_data)

        # Trim trailing /r /n
        credentials.token = credentials.token.strip()

        if is_token_expired(credentials.token):
            print("Credentials expired. Logging in again...")
            delete_credentials()
            return empty_credentials

        return credentials
    except Exception as e:
        dbos_logger.error(f"Error loading credentials: {str(e)}")
        return empty_credentials


def get_cloud_credentials() -> DBOSCloudCredentials:
    # Check if credentials exist and are not expired
    credentials = check_credentials()

    # Log in the user
    if not credentials.token:
        auth_response = authenticate()
        if auth_response is None:
            dbos_logger.error("Failed to login. Exiting...")
            exit(1)
        credentials.token = auth_response.token
        credentials.refreshToken = auth_response.refresh_token
        write_credentials(credentials)

    # Check if the user exists in DBOS Cloud
    user_exists = check_user_profile(credentials)
    if user_exists:
        write_credentials(credentials)
        print(f"[green]Successfully logged in as {credentials.userName}[/green]")
        return credentials

    # User doesn't exist, register the user in DBOS Cloud
    register_user(credentials)
    write_credentials(credentials)

    return credentials



import json
import os
import time
from typing import TYPE_CHECKING, Optional, TypedDict, cast

import docker  # type: ignore
import typer
import yaml
from rich import print
from sqlalchemy import URL, create_engine, text

if TYPE_CHECKING:
    from ._dbos_config import ConfigFile

from ._cloudutils.cloudutils import get_cloud_credentials
from ._cloudutils.databases import choose_database, get_user_db_credentials
from ._error import DBOSInitializationError
from ._logger import dbos_logger

DB_CONNECTION_PATH = os.path.join(".dbos", "db_connection")


class DatabaseConnection(TypedDict):
    hostname: Optional[str]
    port: Optional[int]
    username: Optional[str]
    password: Optional[str]
    local_suffix: Optional[bool]


def db_wizard(config: "ConfigFile", config_file_path: str) -> "ConfigFile":
    # 1. Check the connectivity to the database. Return if successful. If cannot connect, continue to the following steps.
    db_connection_error = _check_db_connectivity(config)
    if db_connection_error is None:
        return config

    # 2. If the error is due to password authentication or the configuration is non-default, surface the error and exit.
    error_str = str(db_connection_error)
    if (
        "password authentication failed" in error_str
        or "28P01" in error_str
        or "no password supplied" in error_str
    ):
        raise DBOSInitializationError(
            f"Could not connect to Postgres: password authentication failed: {db_connection_error}"
        )
    db_config = config["database"]

    # Read the config file and check if the database hostname/port/username are set. If so, skip the wizard.
    with open(config_file_path, "r") as file:
        content = file.read()
        local_config = yaml.safe_load(content)
        if "database" not in local_config:
            local_config["database"] = {}
        local_config = cast("ConfigFile", local_config)

    if (
        local_config["database"].get("hostname")
        or local_config["database"].get("port")
        or local_config["database"].get("username")
        or db_config["hostname"] != "localhost"
        or db_config["port"] != 5432
        or db_config["username"] != "postgres"
    ):
        raise DBOSInitializationError(
            f"Could not connect to the database. Exception: {db_connection_error}"
        )
    print("[yellow]Postgres not detected locally[/yellow]")

    # 3. If the database config is the default one, check if the user has Docker properly installed.
    print("Attempting to start Postgres via Docker")
    has_docker = _check_docker_installed()

    # 4. If Docker is installed, prompt the user to start a local Docker based Postgres, and then set the PGPASSWORD to 'dbos' and try to connect to the database.
    docker_started = False
    if has_docker:
        docker_started = _start_docker_postgres(config)
    else:
        print("[yellow]Docker not detected locally[/yellow]")

    # 5. If no Docker, then prompt the user to log in to DBOS Cloud and provision a DB there. Wait for the remote DB to be ready, and then create a copy of the original config file, and then load the remote connection string to the local config file.
    if not docker_started:
        print("Attempting to connect to Postgres via DBOS Cloud")
        cred = get_cloud_credentials()
        db = choose_database(cred)
        if db is None:
            raise DBOSInitializationError("Error connecting to cloud database")
        config["database"]["hostname"] = db.HostName
        config["database"]["port"] = db.Port
        if db.SupabaseReference is not None:
            config["database"]["username"] = f"postgres.{db.SupabaseReference}"
            supabase_password = typer.prompt(
                "Enter your Supabase database password", hide_input=True
            )
            config["database"]["password"] = supabase_password
        else:
            config["database"]["username"] = db.DatabaseUsername
            db_credentials = get_user_db_credentials(cred, db.PostgresInstanceName)
            config["database"]["password"] = db_credentials.Password
        config["database"]["local_suffix"] = True

        # Verify these new credentials work
        db_connection_error = _check_db_connectivity(config)
        if db_connection_error is not None:
            raise DBOSInitializationError(
                f"Could not connect to the database. Exception: {db_connection_error}"
            )

    # 6. Save the config to the database connection file
    updated_connection = DatabaseConnection(
        hostname=config["database"]["hostname"],
        port=config["database"]["port"],
        username=config["database"]["username"],
        password=config["database"]["password"],
        local_suffix=config["database"]["local_suffix"],
    )
    save_db_connection(updated_connection)
    return config


def _start_docker_postgres(config: "ConfigFile") -> bool:
    print("Starting a Postgres Docker container...")
    client = docker.from_env()
    pg_data = "/var/lib/postgresql/data"
    container_name = "dbos-db"
    client.containers.run(
        image="pgvector/pgvector:pg16",
        detach=True,
        environment={
            "POSTGRES_PASSWORD": config["database"]["password"],
            "PGDATA": pg_data,
        },
        volumes={pg_data: {"bind": pg_data, "mode": "rw"}},
        ports={"5432/tcp": config["database"]["port"]},
        name=container_name,
        remove=True,
    )

    container = client.containers.get(container_name)
    attempts = 30
    while attempts > 0:
        if attempts % 5 == 0:
            print("Waiting for Postgres Docker container to start...")
        try:
            res = container.exec_run("psql -U postgres -c 'SELECT 1;'")
            if res.exit_code != 0:
                attempts -= 1
                time.sleep(1)
                continue
            print("[green]Postgres Docker container started successfully![/green]")
            break
        except:
            attempts -= 1
            time.sleep(1)

    if attempts == 0:
        print("[yellow]Failed to start Postgres Docker container.[/yellow]")
        return False

    return True


def _check_docker_installed() -> bool:
    # Check if Docker is installed
    try:
        client = docker.from_env()
        client.ping()
    except Exception:
        return False
    return True


def _check_db_connectivity(config: "ConfigFile") -> Optional[Exception]:
    postgres_db_url = URL.create(
        "postgresql+psycopg",
        username=config["database"]["username"],
        password=config["database"]["password"],
        host=config["database"]["hostname"],
        port=config["database"]["port"],
        database="postgres",
        query={"connect_timeout": "1"},
    )
    postgres_db_engine = create_engine(postgres_db_url)
    try:
        with postgres_db_engine.connect() as conn:
            val = conn.execute(text("SELECT 1")).scalar()
            if val != 1:
                dbos_logger.error(
                    f"Unexpected value returned from database: expected 1, received {val}"
                )
                return Exception()
    except Exception as e:
        return e
    finally:
        postgres_db_engine.dispose()

    return None


def load_db_connection() -> DatabaseConnection:
    try:
        with open(DB_CONNECTION_PATH, "r") as f:
            data = json.load(f)
            return DatabaseConnection(
                hostname=data.get("hostname", None),
                port=data.get("port", None),
                username=data.get("username", None),
                password=data.get("password", None),
                local_suffix=data.get("local_suffix", None),
            )
    except:
        return DatabaseConnection(
            hostname=None, port=None, username=None, password=None, local_suffix=None
        )


def save_db_connection(connection: DatabaseConnection) -> None:
    os.makedirs(".dbos", exist_ok=True)
    with open(DB_CONNECTION_PATH, "w") as f:
        json.dump(connection, f)



# Welcome to DBOS!

# This is a sample app built with DBOS and FastAPI.
# It displays greetings to visitors and keeps track of how
# many times visitors have been greeted.

# First, let's do imports, create a FastAPI app, and initialize DBOS.

from fastapi import FastAPI
from fastapi.responses import HTMLResponse

from dbos import DBOS

from .schema import dbos_hello

app = FastAPI()
DBOS(fastapi=app)

# Next, let's write a function that greets visitors.
# To make it more interesting, we'll keep track of how
# many times visitors have been greeted and store
# the count in the database.

# We implement the database operations using SQLAlchemy
# and serve the function from a FastAPI endpoint.
# We annotate it with @DBOS.transaction() to access
# an automatically-configured database client.


@app.get("/greeting/{name}")
@DBOS.transaction()
def example_transaction(name: str) -> str:
    query = dbos_hello.insert().values(name=name).returning(dbos_hello.c.greet_count)
    greet_count = DBOS.sql_session.execute(query).scalar_one()
    greeting = f"Greetings, {name}! You have been greeted {greet_count} times."
    DBOS.logger.info(greeting)
    return greeting


# Finally, let's use FastAPI to serve an HTML + CSS readme
# from the root path.


@app.get("/")
def readme() -> HTMLResponse:
    readme = """
        <!DOCTYPE html>
        <html lang="en">
        <head>
            <title>Welcome to DBOS!</title>
            <script src="https://cdn.tailwindcss.com"></script>
        </head>
        <body class="font-sans text-gray-800 p-6 max-w-2xl mx-auto">
            <h1 class="text-xl font-semibold mb-4">Welcome to DBOS!</h1>
            <p class="mb-4">
                Visit the route <code class="bg-gray-100 px-1 rounded">/greeting/{name}</code> to be greeted!<br>
                For example, visit <code class="bg-gray-100 px-1 rounded"><a href="/greeting/dbos" class="text-blue-600 hover:underline">/greeting/dbos</a></code><br>
                The counter increments with each page visit.<br>
            </p>
            <p>
                To learn more about DBOS, check out the <a href="https://docs.dbos.dev" class="text-blue-600 hover:underline">docs</a>.
            </p>
        </body>
        </html>
        """
    return HTMLResponse(readme)


# To deploy this app to DBOS Cloud:
# - "npm i -g @dbos-inc/dbos-cloud@latest" to install the Cloud CLI (requires Node)
# - "dbos-cloud app deploy" to deploy your app
# - Deploy outputs a URL--visit it to see your app!


# To run this app locally:
# - Make sure you have a Postgres database to connect to
# - "dbos migrate" to set up your database tables
# - "dbos start" to start the app
# - Visit localhost:8000 to see your app!



import time
from datetime import datetime

import pytest
from sqlalchemy import create_engine, text
from sqlalchemy.engine import Engine

# Public API
from dbos import DBOS


def simulate_db_restart(engine: Engine, downtime: float) -> None:
    # Get DB name
    with engine.connect() as connection:
        current_db = connection.execute(text("SELECT current_database()")).scalar()

    # Goal here is to disable connections to the DB for a while.
    #   Need a temp DB to do that and recover connectivity...
    temp_db_name = "temp_database_for_maintenance"

    # Retrieve the URL of the current engine
    main_db_url = engine.url

    # Modify the URL to point to the temporary database
    temp_db_url = main_db_url.set(database=temp_db_name)

    try:
        with engine.connect().execution_options(
            isolation_level="AUTOCOMMIT"
        ) as connection:
            # Create a temporary database
            connection.execute(text(f"CREATE DATABASE {temp_db_name};"))
            print(f"Temporary database '{temp_db_name}' created.")
    except Exception as e:
        print("Could not create temp db: ", e)

    temp_engine = create_engine(temp_db_url)
    with temp_engine.connect().execution_options(
        isolation_level="AUTOCOMMIT"
    ) as temp_connection:
        try:
            # Disable new connections to the database
            temp_connection.execute(
                text(f"ALTER DATABASE {current_db} WITH ALLOW_CONNECTIONS false;")
            )

            # Terminate all connections except the current one
            temp_connection.execute(
                text(
                    f"""
                SELECT pg_terminate_backend(pid)
                FROM pg_stat_activity
                WHERE pid <> pg_backend_pid()
                AND datname = '{current_db}';
            """
                )
            )
        except Exception as e:
            print(f"Could not disable db {current_db}: ", e)

        time.sleep(downtime)

        # Re-enable new connections
        try:
            temp_connection.execute(
                text(f"ALTER DATABASE {current_db} WITH ALLOW_CONNECTIONS true;")
            )
        except Exception as e:
            print(f"Could not reenable db {current_db}: ", e)
    temp_engine.dispose()

    try:
        with engine.connect().execution_options(
            isolation_level="AUTOCOMMIT"
        ) as connection:
            # Clean up the temp DB
            connection.execute(
                text(
                    f"""
                SELECT pg_terminate_backend(pid)
                FROM pg_stat_activity
                WHERE pid <> pg_backend_pid()
                AND datname = '{temp_db_name}';
            """
                )
            )

            # Drop temporary database
            connection.execute(text(f"DROP DATABASE {temp_db_name};"))
    except Exception as e:
        print(f"Could not clean up temp db {temp_db_name}: ", e)


def test_scheduled_workflow(dbos: DBOS) -> None:
    wf_counter: int = 0

    @DBOS.scheduled("* * * * * *")
    @DBOS.workflow()
    def test_workflow(scheduled: datetime, actual: datetime) -> None:
        nonlocal wf_counter
        wf_counter += 1

    time.sleep(4)
    assert wf_counter > 2 and wf_counter <= 4


def test_appdb_downtime(dbos: DBOS) -> None:
    wf_counter: int = 0

    @DBOS.transaction()
    def test_transaction(var2: str) -> str:
        rows = DBOS.sql_session.execute(text("SELECT 1")).fetchall()
        return "ran"

    @DBOS.scheduled("* * * * * *")
    @DBOS.workflow()
    def test_workflow(scheduled: datetime, actual: datetime) -> None:
        nonlocal wf_counter
        test_transaction("x")
        wf_counter += 1

    time.sleep(2)
    simulate_db_restart(dbos._app_db.engine, 2)
    time.sleep(2)
    assert wf_counter > 2


def test_sysdb_downtime(dbos: DBOS) -> None:
    wf_counter: int = 0

    @DBOS.scheduled("* * * * * *")
    @DBOS.workflow()
    def test_workflow(scheduled: datetime, actual: datetime) -> None:
        nonlocal wf_counter
        wf_counter += 1

    time.sleep(2)
    simulate_db_restart(dbos._sys_db.engine, 2)
    time.sleep(2)
    # We know there should be at least 3 occurrences from the 4 seconds when the DB was up.
    #  There could be more than 4, depending on the pace the machine...
    assert wf_counter > 2


def test_scheduled_transaction(dbos: DBOS) -> None:
    txn_counter: int = 0

    @DBOS.scheduled("* * * * * *")
    @DBOS.transaction()
    def test_transaction(scheduled: datetime, actual: datetime) -> None:
        nonlocal txn_counter
        txn_counter += 1

    time.sleep(4)
    assert txn_counter > 2 and txn_counter <= 4


def test_scheduled_workflow_exception(dbos: DBOS) -> None:
    wf_counter: int = 0

    @DBOS.scheduled("* * * * * *")
    @DBOS.workflow()
    def test_failing_workflow(scheduled: datetime, actual: datetime) -> None:
        nonlocal wf_counter
        wf_counter += 1
        raise Exception("error")

    time.sleep(3)
    assert wf_counter >= 1 and wf_counter <= 3


def test_scheduler_oaoo(dbos: DBOS) -> None:
    wf_counter: int = 0
    txn_counter: int = 0
    workflow_id: str = ""

    @DBOS.scheduled("* * * * * *")
    @DBOS.workflow()
    def test_workflow(scheduled: datetime, actual: datetime) -> None:
        test_transaction()
        nonlocal wf_counter
        wf_counter += 1
        nonlocal workflow_id
        workflow_id = DBOS.workflow_id

    @DBOS.transaction()
    def test_transaction() -> None:
        nonlocal txn_counter
        txn_counter += 1

    time.sleep(3)
    assert wf_counter >= 1 and wf_counter <= 3
    max_tries = 10
    for i in range(max_tries):
        try:
            assert txn_counter == wf_counter
            break
        except Exception as e:
            if i == max_tries - 1:
                raise e
            else:
                time.sleep(1)

    # Stop all scheduled workflows
    for event in dbos.stop_events:
        event.set()

    dbos._sys_db.wait_for_buffer_flush()
    dbos._sys_db.update_workflow_status(
        {
            "workflow_uuid": workflow_id,
            "status": "PENDING",
            "name": test_workflow.__qualname__,
            "class_name": None,
            "config_name": None,
            "output": None,
            "error": None,
            "executor_id": None,
            "app_id": None,
            "app_version": None,
            "request": None,
            "recovery_attempts": None,
            "authenticated_user": None,
            "authenticated_roles": None,
            "assumed_role": None,
            "queue_name": None,
            "created_at": None,
            "updated_at": None,
        }
    )

    workflow_handles = DBOS.recover_pending_workflows()
    assert len(workflow_handles) == 1
    assert workflow_handles[0].get_result() == None
    max_tries = 10
    for i in range(max_tries):
        try:
            assert txn_counter + 1 == wf_counter
            break
        except Exception as e:
            if i == max_tries - 1:
                raise e
            else:
                time.sleep(1)


def test_long_workflow(dbos: DBOS) -> None:
    """
    This runs every hour and does nothing. Goal is to verify that it shuts down properly.
    """

    @DBOS.scheduled("0 * * * *")
    @DBOS.workflow()
    def test_workflow(scheduled: datetime, actual: datetime) -> None:
        pass


def test_bad_schedule(dbos: DBOS) -> None:
    code = """
@DBOS.scheduled("*/invalid * * * * *")  # this is not a valid schedule
def my_function():
    pass
"""
    # Use exec to run the code and catch the expected exception
    with pytest.raises(ValueError, match="Invalid crontab"):
        exec(code)



import os
import platform
import signal
import subprocess
import time
import typing
from os import path
from typing import Any

import jsonpickle  # type: ignore
import requests
import sqlalchemy as sa
import typer
from rich import print
from rich.prompt import IntPrompt
from typing_extensions import Annotated

from dbos._debug import debug_workflow, parse_start_command

from .. import load_config
from .._app_db import ApplicationDatabase
from .._dbos_config import _is_valid_app_name
from .._sys_db import SystemDatabase, reset_system_database
from .._workflow_commands import get_workflow, list_queued_workflows, list_workflows
from ..cli._github_init import create_template_from_github
from ._template_init import copy_template, get_project_name, get_templates_directory

app = typer.Typer()
workflow = typer.Typer()
queue = typer.Typer()

app.add_typer(workflow, name="workflow", help="Manage DBOS workflows")
workflow.add_typer(queue, name="queue", help="Manage enqueued workflows")


def _on_windows() -> bool:
    return platform.system() == "Windows"


@app.command(
    help="Start your DBOS application using the start commands in 'dbos-config.yaml'"
)
def start() -> None:
    config = load_config()
    start_commands = config["runtimeConfig"]["start"]
    typer.echo("Executing start commands from 'dbos-config.yaml'")
    for command in start_commands:
        typer.echo(f"Executing: {command}")

        # Run the command in the child process.
        # On Unix-like systems, set its process group
        process = subprocess.Popen(
            command,
            shell=True,
            text=True,
            preexec_fn=os.setsid if not _on_windows() else None,
        )

        def signal_handler(signum: int, frame: Any) -> None:
            """
            Forward kill signals to children.

            When we receive a signal, send it to the entire process group of the child.
            If that doesn't work, SIGKILL them then exit.
            """
            # Send the signal to the child's entire process group
            if process.poll() is None:
                os.killpg(os.getpgid(process.pid), signum)

            # Give some time for the child to terminate
            for _ in range(10):  # Wait up to 1 second
                if process.poll() is not None:
                    break
                time.sleep(0.1)

            # If the child is still running, force kill it
            if process.poll() is None:
                try:
                    os.killpg(os.getpgid(process.pid), signal.SIGKILL)
                except Exception:
                    pass

            # Exit immediately
            os._exit(process.returncode if process.returncode is not None else 1)

        # Configure the single handler only on Unix-like systems.
        # TODO: Also kill the children on Windows.
        if not _on_windows():
            signal.signal(signal.SIGINT, signal_handler)
            signal.signal(signal.SIGTERM, signal_handler)
        process.wait()


@app.command(help="Initialize a new DBOS application from a template")
def init(
    project_name: Annotated[
        typing.Optional[str], typer.Argument(help="Specify application name")
    ] = None,
    template: Annotated[
        typing.Optional[str],
        typer.Option("--template", "-t", help="Specify template to use"),
    ] = None,
    config: Annotated[
        bool,
        typer.Option("--config", "-c", help="Only add dbos-config.yaml"),
    ] = False,
) -> None:
    try:

        git_templates = ["dbos-toolbox", "dbos-app-starter", "dbos-cron-starter"]
        templates_dir = get_templates_directory()
        templates = git_templates + [
            x.name for x in os.scandir(templates_dir) if x.is_dir()
        ]

        if config and template is None:
            template = templates[-1]

        if template:
            if template not in templates:
                raise Exception(f"Template {template} not found in {templates_dir}")
        else:
            print("\n[bold]Available templates:[/bold]")
            for idx, template_name in enumerate(templates, 1):
                print(f"  {idx}. {template_name}")
            while True:
                try:
                    choice = IntPrompt.ask(
                        "\nSelect template number",
                        show_choices=False,
                        show_default=False,
                    )
                    if 1 <= choice <= len(templates):
                        template = templates[choice - 1]
                        break
                    else:
                        print(
                            "[red]Invalid selection. Please choose a number from the list.[/red]"
                        )
                except (KeyboardInterrupt, EOFError):
                    raise typer.Abort()
                except ValueError:
                    print("[red]Please enter a valid number.[/red]")

        if template in git_templates:
            project_name = template
        else:
            if project_name is None:
                project_name = typing.cast(
                    str,
                    typer.prompt("What is your project's name?", get_project_name()),
                )

        if not _is_valid_app_name(project_name):
            raise Exception(
                f"{project_name} is an invalid DBOS app name. App names must be between 3 and 30 characters long and contain only lowercase letters, numbers, dashes, and underscores."
            )

        if template in git_templates:
            create_template_from_github(app_name=project_name, template_name=template)
        else:
            copy_template(
                path.join(templates_dir, template), project_name, config_mode=config
            )
    except Exception as e:
        print(f"[red]{e}[/red]")


@app.command(
    help="Run your database schema migrations using the migration commands in 'dbos-config.yaml'"
)
def migrate() -> None:
    config = load_config()
    if not config["database"]["password"]:
        typer.echo(
            "DBOS configuration does not contain database password, please check your config file and retry!"
        )
        raise typer.Exit(code=1)
    app_db_name = config["database"]["app_db_name"]

    typer.echo(f"Starting schema migration for database {app_db_name}")

    # First, run DBOS migrations on the system database and the application database
    app_db = None
    sys_db = None
    try:
        sys_db = SystemDatabase(config)
        app_db = ApplicationDatabase(config)
    except Exception as e:
        typer.echo(f"DBOS system schema migration failed: {e}")
    finally:
        if sys_db:
            sys_db.destroy()
        if app_db:
            app_db.destroy()

    # Next, run any custom migration commands specified in the configuration
    typer.echo("Executing migration commands from 'dbos-config.yaml'")
    try:
        migrate_commands = (
            config["database"]["migrate"]
            if "migrate" in config["database"] and config["database"]["migrate"]
            else []
        )
        for command in migrate_commands:
            typer.echo(f"Executing migration command: {command}")
            result = subprocess.run(command, shell=True, text=True)
            if result.returncode != 0:
                typer.echo(f"Migration command failed: {command}")
                typer.echo(result.stderr)
                raise typer.Exit(1)
            if result.stdout:
                typer.echo(result.stdout.rstrip())
    except Exception as e:
        typer.echo(f"An error occurred during schema migration: {e}")
        raise typer.Exit(code=1)

    typer.echo(f"Completed schema migration for database {app_db_name}")


@app.command(help="Reset the DBOS system database")
def reset(
    yes: bool = typer.Option(False, "-y", "--yes", help="Skip confirmation prompt")
) -> None:
    if not yes:
        confirm = typer.confirm(
            "This command resets your DBOS system database, deleting metadata about past workflows and steps. Are you sure you want to proceed?"
        )
        if not confirm:
            typer.echo("Operation cancelled.")
            raise typer.Exit()
    config = load_config()
    try:
        reset_system_database(config)
    except sa.exc.SQLAlchemyError as e:
        typer.echo(f"Error resetting system database: {str(e)}")
        return


@app.command(help="Replay Debug a DBOS workflow")
def debug(
    workflow_id: Annotated[str, typer.Argument(help="Workflow ID to debug")],
) -> None:
    config = load_config(silent=True, use_db_wizard=False)
    start = config["runtimeConfig"]["start"]
    if not start:
        typer.echo("No start commands found in 'dbos-config.yaml'")
        raise typer.Exit(code=1)
    if len(start) > 1:
        typer.echo("Multiple start commands found in 'dbos-config.yaml'")
        raise typer.Exit(code=1)
    entrypoint = parse_start_command(start[0])
    debug_workflow(workflow_id, entrypoint)


@workflow.command(help="List workflows for your application")
def list(
    limit: Annotated[
        int,
        typer.Option("--limit", "-l", help="Limit the results returned"),
    ] = 10,
    user: Annotated[
        typing.Optional[str],
        typer.Option("--user", "-u", help="Retrieve workflows run by this user"),
    ] = None,
    starttime: Annotated[
        typing.Optional[str],
        typer.Option(
            "--start-time",
            "-s",
            help="Retrieve workflows starting after this timestamp (ISO 8601 format)",
        ),
    ] = None,
    endtime: Annotated[
        typing.Optional[str],
        typer.Option(
            "--end-time",
            "-e",
            help="Retrieve workflows starting before this timestamp (ISO 8601 format)",
        ),
    ] = None,
    status: Annotated[
        typing.Optional[str],
        typer.Option(
            "--status",
            "-S",
            help="Retrieve workflows with this status (PENDING, SUCCESS, ERROR, RETRIES_EXCEEDED, ENQUEUED, or CANCELLED)",
        ),
    ] = None,
    appversion: Annotated[
        typing.Optional[str],
        typer.Option(
            "--application-version",
            "-v",
            help="Retrieve workflows with this application version",
        ),
    ] = None,
    name: Annotated[
        typing.Optional[str],
        typer.Option(
            "--name",
            "-n",
            help="Retrieve workflows with this name",
        ),
    ] = None,
    request: Annotated[
        bool,
        typer.Option("--request", help="Retrieve workflow request information"),
    ] = False,
) -> None:
    config = load_config(silent=True)
    sys_db = SystemDatabase(config)
    workflows = list_workflows(
        sys_db,
        limit=limit,
        user=user,
        start_time=starttime,
        end_time=endtime,
        status=status,
        request=request,
        app_version=appversion,
        name=name,
    )
    print(jsonpickle.encode(workflows, unpicklable=False))


@workflow.command(help="Retrieve the status of a workflow")
def get(
    workflow_id: Annotated[str, typer.Argument()],
    request: Annotated[
        bool,
        typer.Option("--request", help="Retrieve workflow request information"),
    ] = False,
) -> None:
    config = load_config(silent=True)
    sys_db = SystemDatabase(config)
    print(
        jsonpickle.encode(get_workflow(sys_db, workflow_id, request), unpicklable=False)
    )


@workflow.command(
    help="Cancel a workflow so it is no longer automatically retried or restarted"
)
def cancel(
    uuid: Annotated[str, typer.Argument()],
    host: Annotated[
        typing.Optional[str],
        typer.Option("--host", "-H", help="Specify the admin host"),
    ] = "localhost",
    port: Annotated[
        typing.Optional[int],
        typer.Option("--port", "-p", help="Specify the admin port"),
    ] = 3001,
) -> None:
    response = requests.post(
        f"http://{host}:{port}/workflows/{uuid}/cancel", json=[], timeout=5
    )

    if response.status_code == 204:
        print(f"Workflow {uuid} has been cancelled")
    else:
        print(f"Failed to cancel workflow {uuid}. Status code: {response.status_code}")


@workflow.command(help="Resume a workflow that has been cancelled")
def resume(
    uuid: Annotated[str, typer.Argument()],
    host: Annotated[
        typing.Optional[str],
        typer.Option("--host", "-H", help="Specify the admin host"),
    ] = "localhost",
    port: Annotated[
        typing.Optional[int],
        typer.Option("--port", "-p", help="Specify the admin port"),
    ] = 3001,
) -> None:
    response = requests.post(
        f"http://{host}:{port}/workflows/{uuid}/resume", json=[], timeout=5
    )

    if response.status_code == 204:
        print(f"Workflow {uuid} has been resumed")
    else:
        print(f"Failed to resume workflow {uuid}. Status code: {response.status_code}")


@workflow.command(help="Restart a workflow from the beginning with a new id")
def restart(
    uuid: Annotated[str, typer.Argument()],
    host: Annotated[
        typing.Optional[str],
        typer.Option("--host", "-H", help="Specify the admin host"),
    ] = "localhost",
    port: Annotated[
        typing.Optional[int],
        typer.Option("--port", "-p", help="Specify the admin port"),
    ] = 3001,
) -> None:
    response = requests.post(
        f"http://{host}:{port}/workflows/{uuid}/restart", json=[], timeout=5
    )

    if response.status_code == 204:
        print(f"Workflow {uuid} has been restarted")
    else:
        print(f"Failed to resume workflow {uuid}. Status code: {response.status_code}")


@queue.command(name="list", help="List enqueued functions for your application")
def list_queue(
    limit: Annotated[
        typing.Optional[int],
        typer.Option("--limit", "-l", help="Limit the results returned"),
    ] = None,
    start_time: Annotated[
        typing.Optional[str],
        typer.Option(
            "--start-time",
            "-s",
            help="Retrieve functions starting after this timestamp (ISO 8601 format)",
        ),
    ] = None,
    end_time: Annotated[
        typing.Optional[str],
        typer.Option(
            "--end-time",
            "-e",
            help="Retrieve functions starting before this timestamp (ISO 8601 format)",
        ),
    ] = None,
    status: Annotated[
        typing.Optional[str],
        typer.Option(
            "--status",
            "-S",
            help="Retrieve functions with this status (PENDING, SUCCESS, ERROR, RETRIES_EXCEEDED, ENQUEUED, or CANCELLED)",
        ),
    ] = None,
    queue_name: Annotated[
        typing.Optional[str],
        typer.Option(
            "--queue-name",
            "-q",
            help="Retrieve functions on this queue",
        ),
    ] = None,
    name: Annotated[
        typing.Optional[str],
        typer.Option(
            "--name",
            "-n",
            help="Retrieve functions on this queue",
        ),
    ] = None,
    request: Annotated[
        bool,
        typer.Option("--request", help="Retrieve workflow request information"),
    ] = False,
) -> None:
    config = load_config(silent=True)
    sys_db = SystemDatabase(config)
    workflows = list_queued_workflows(
        sys_db=sys_db,
        limit=limit,
        start_time=start_time,
        end_time=end_time,
        queue_name=queue_name,
        status=status,
        request=request,
        name=name,
    )
    print(jsonpickle.encode(workflows, unpicklable=False))


if __name__ == "__main__":
    app()



import re
import sys
from typing import NoReturn, Optional, Union

from dbos.cli.cli import app


def main() -> NoReturn:
    # Modify sys.argv[0] to remove script or executable extensions
    sys.argv[0] = re.sub(r"(-script\.pyw|\.exe)?$", "", sys.argv[0])

    retval: Optional[Union[str, int]] = 1
    try:
        app()
        retval = None
    except SystemExit as e:
        retval = e.code
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        retval = 1
    finally:
        sys.exit(retval)


if __name__ == "__main__":
    main()



import logging
import uuid
from typing import Tuple

import pytest
import sqlalchemy as sa
from flask import Flask, Response, jsonify

from dbos import DBOS
from dbos._context import assert_current_dbos_context


def test_flask_endpoint(
    dbos_flask: Tuple[DBOS, Flask], caplog: pytest.LogCaptureFixture
) -> None:
    _, app = dbos_flask

    @app.route("/endpoint/<var1>/<var2>")
    def test_endpoint(var1: str, var2: str) -> Response:
        ctx = assert_current_dbos_context()
        assert not ctx.is_within_workflow()
        return test_workflow(var1, var2)

    @app.route("/workflow/<var1>/<var2>")
    @DBOS.workflow()
    def test_workflow(var1: str, var2: str) -> Response:
        assert DBOS.request is not None
        assert DBOS.request.headers.get("Host") == "localhost"
        res1 = test_transaction(var1)
        res2 = test_step(var2)
        result = res1 + res2
        return jsonify({"result": result})

    @app.route("/transaction/<var>")
    @DBOS.transaction()
    def test_transaction(var: str) -> str:
        rows = DBOS.sql_session.execute(sa.text("SELECT 1")).fetchall()
        return var + str(rows[0][0])

    @DBOS.step()
    def test_step(var: str) -> str:
        return var

    app.config["TESTING"] = True
    client = app.test_client()

    original_propagate = logging.getLogger("dbos").propagate
    caplog.set_level(logging.WARNING, "dbos")
    logging.getLogger("dbos").propagate = True

    response = client.get("/endpoint/a/b")
    assert response.status_code == 200
    assert response.json == {"result": "a1b"}
    assert caplog.text == ""

    response = client.get("/workflow/a/b")
    assert response.status_code == 200
    assert response.json == {"result": "a1b"}
    assert caplog.text == ""

    response = client.get("/transaction/bob")
    assert response.status_code == 200
    assert response.text == "bob1"
    assert caplog.text == ""

    # Reset logging
    logging.getLogger("dbos").propagate = original_propagate


def test_endpoint_recovery(dbos_flask: Tuple[DBOS, Flask]) -> None:
    dbos, app = dbos_flask

    wfuuid = str(uuid.uuid4())

    @DBOS.workflow()
    def test_workflow(var1: str) -> tuple[str, str]:
        assert DBOS.request is not None
        return var1, DBOS.workflow_id

    @app.route("/<var1>/<var2>")
    def test_endpoint(var1: str, var2: str) -> dict[str, str]:
        assert DBOS.request is not None
        res1, id1 = test_workflow(var1)
        res2, id2 = test_workflow(var2)
        return {"res1": res1, "res2": res2, "id1": id1, "id2": id2}

    app.config["TESTING"] = True
    client = app.test_client()

    response = client.get("/a/b", headers={"dbos-idempotency-key": wfuuid})
    assert response.status_code == 200
    assert response.json is not None
    assert response.json.get("res1") == "a"
    assert response.json.get("res2") == "b"
    assert response.json.get("id1") == wfuuid
    assert response.json.get("id2") != wfuuid

    dbos._sys_db.wait_for_buffer_flush()
    # Change the workflow status to pending
    dbos._sys_db.update_workflow_status(
        {
            "workflow_uuid": wfuuid,
            "status": "PENDING",
            "name": test_workflow.__qualname__,
            "class_name": None,
            "config_name": None,
            "output": None,
            "error": None,
            "executor_id": None,
            "app_id": None,
            "app_version": None,
            "request": None,
            "recovery_attempts": None,
            "authenticated_user": None,
            "authenticated_roles": None,
            "assumed_role": None,
            "queue_name": None,
            "created_at": None,
            "updated_at": None,
        }
    )

    # Recovery should execute the workflow again but skip the transaction
    workflow_handles = DBOS.recover_pending_workflows()
    assert len(workflow_handles) == 1
    assert workflow_handles[0].get_result() == ("a", wfuuid)



from dbos import DBOS
from dbos._dbos_config import ConfigFile


@DBOS.workflow()
def my_function(foo: str) -> str:
    return foo


def default_config() -> ConfigFile:
    return {
        "name": "forgot-launch",
        "language": "python",
        "database": {
            "hostname": "localhost",
            "port": 5432,
            "username": "postgres",
            "password": "doesntmatter",
            "app_db_name": "doesntmatter",
        },
        "runtimeConfig": {
            "start": ["doesntmatter"],
        },
        "telemetry": {},
        "env": {},
    }


DBOS(config=default_config())



from dataclasses import dataclass
from typing import Optional, Union


@dataclass
class KafkaMessage:
    headers: Optional[list[tuple[str, Union[str, bytes]]]]
    key: Optional[Union[str, bytes]]
    latency: Optional[float]
    leader_epoch: Optional[int]
    offset: Optional[int]
    partition: Optional[int]
    timestamp: tuple[int, int]
    topic: Optional[str]
    value: Optional[Union[str, bytes]]



from typing import Optional

# Public API
from dbos import DBOS


@DBOS.dbos_class()
class DBOSWFEvents:
    @staticmethod
    @DBOS.workflow()
    def test_setevent_workflow() -> None:
        DBOS.set_event("key1", "value1")
        DBOS.set_event("key2", "value2")
        DBOS.set_event("key3", None)

    @staticmethod
    @DBOS.workflow()
    def test_getevent_workflow(
        target_uuid: str, key: str, timeout_seconds: float = 10
    ) -> Optional[str]:
        msg = DBOS.get_event(target_uuid, key, timeout_seconds)
        return str(msg) if msg is not None else None


@DBOS.workflow()
def wfFunc(arg: str) -> str:
    assert DBOS.workflow_id == "wfid"
    return arg + "1"



import os
from base64 import b64decode
from typing import List, TypedDict

import requests

DEMO_REPO_API = "https://api.github.com/repos/dbos-inc/dbos-demo-apps"
PY_DEMO_PATH = "python/"
BRANCH = "main"


class GitHubTreeItem(TypedDict):
    path: str
    mode: str
    type: str
    sha: str
    url: str
    size: int


class GitHubTree(TypedDict):
    sha: str
    url: str
    tree: List[GitHubTreeItem]
    truncated: bool


class GitHubItem(TypedDict):
    sha: str
    node_id: str
    url: str
    content: str
    encoding: str
    size: int


def _fetch_github(url: str) -> requests.Response:
    headers = {}
    github_token = os.getenv("GITHUB_TOKEN")
    if github_token:
        headers["Authorization"] = f"Bearer {github_token}"

    response = requests.get(url, headers=headers)

    if not response.ok:
        if response.headers.get("x-ratelimit-remaining") == "0":
            raise Exception(
                "Error fetching from GitHub API: rate limit exceeded.\n"
                "Please wait a few minutes and try again.\n"
                "To increase the limit, you can create a personal access token and set it in the GITHUB_TOKEN environment variable.\n"
                "Details: https://docs.github.com/en/rest/using-the-rest-api/rate-limits-for-the-rest-api"
            )
        elif response.status_code == 401:
            raise Exception(
                f"Error fetching content from GitHub {url}: {response.status_code} {response.reason}.\n"
                "Please ensure your GITHUB_TOKEN environment variable is set to a valid personal access token."
            )
        raise Exception(
            f"Error fetching content from GitHub {url}: {response.status_code} {response.reason}"
        )

    return response


def _fetch_github_tree(tag: str) -> List[GitHubTreeItem]:
    response = _fetch_github(f"{DEMO_REPO_API}/git/trees/{tag}?recursive=1")
    tree_data: GitHubTree = response.json()
    return tree_data["tree"]


def _fetch_github_item(url: str) -> str:
    response = _fetch_github(url)
    item: GitHubItem = response.json()
    return b64decode(item["content"]).decode("utf-8")


def create_template_from_github(app_name: str, template_name: str) -> None:
    print(
        f"Creating a new application named {app_name} from the template {template_name}"
    )

    tree = _fetch_github_tree(BRANCH)
    template_path = f"{PY_DEMO_PATH}{template_name}/"

    files_to_download = [
        item
        for item in tree
        if item["path"].startswith(template_path) and item["type"] == "blob"
    ]

    # Download every file from the template
    for item in files_to_download:
        raw_content = _fetch_github_item(item["url"])
        file_path = item["path"].replace(template_path, "")
        target_path = os.path.join(".", file_path)

        # Create directory if it doesn't exist
        os.makedirs(os.path.dirname(target_path), exist_ok=True)

        # Write file with proper permissions
        with open(target_path, "w", encoding="utf-8") as f:
            f.write(raw_content)
        os.chmod(target_path, int(item["mode"], 8))

    print(
        f"Downloaded {len(files_to_download)} files from the template GitHub repository"
    )



import asyncio
import logging
import uuid
from contextlib import asynccontextmanager
from typing import Any, Tuple

import httpx
import pytest
import sqlalchemy as sa
import uvicorn
from fastapi import FastAPI
from fastapi.testclient import TestClient

# Public API
from dbos import DBOS, ConfigFile

# Private API because this is a unit test
from dbos._context import assert_current_dbos_context


def test_simple_endpoint(
    dbos_fastapi: Tuple[DBOS, FastAPI], caplog: pytest.LogCaptureFixture
) -> None:
    dbos, app = dbos_fastapi
    client = TestClient(app)

    @app.get("/endpoint/{var1}/{var2}")
    def test_endpoint(var1: str, var2: str) -> str:
        result = test_workflow(var1, var2)
        ctx = assert_current_dbos_context()
        assert not ctx.is_within_workflow()
        return result

    @app.get("/workflow/{var1}/{var2}")
    @DBOS.workflow()
    def test_workflow(var1: str, var2: str) -> str:
        DBOS.span.set_attribute("test_key", "test_value")
        assert DBOS.request is not None
        res1 = test_transaction(var1)
        res2 = test_step(var2)
        return res1 + res2

    @app.get("/transaction/{var}")
    @DBOS.transaction()
    def test_transaction(var: str) -> str:
        rows = DBOS.sql_session.execute(sa.text("SELECT 1")).fetchall()
        return var + str(rows[0][0])

    @DBOS.step()
    def test_step(var: str) -> str:
        return var

    original_propagate = logging.getLogger("dbos").propagate
    caplog.set_level(logging.WARNING, "dbos")
    logging.getLogger("dbos").propagate = True

    response = client.get("/workflow/bob/bob")
    assert response.status_code == 200
    assert response.text == '"bob1bob"'
    assert caplog.text == ""

    response = client.get("/endpoint/bob/bob")
    assert response.status_code == 200
    assert response.text == '"bob1bob"'
    assert caplog.text == ""

    response = client.get("/transaction/bob")
    assert response.status_code == 200
    assert response.text == '"bob1"'
    assert caplog.text == ""

    # Reset logging
    logging.getLogger("dbos").propagate = original_propagate


def test_start_workflow(dbos_fastapi: Tuple[DBOS, FastAPI]) -> None:
    dbos, app = dbos_fastapi
    client = TestClient(app)

    @app.get("/{var1}/{var2}")
    def test_endpoint(var1: str, var2: str) -> str:
        handle = dbos.start_workflow(test_workflow, var1, var2)
        context = assert_current_dbos_context()
        assert not context.is_within_workflow()
        return handle.get_result()

    @DBOS.workflow()
    def test_workflow(var1: str, var2: str) -> str:
        DBOS.span.set_attribute("test_key", "test_value")
        assert DBOS.request is not None
        res1 = test_transaction(var1)
        res2 = test_step(var2)
        return res1 + res2

    @DBOS.transaction()
    def test_transaction(var: str) -> str:
        rows = DBOS.sql_session.execute(sa.text("SELECT 1")).fetchall()
        return var + str(rows[0][0])

    @DBOS.step()
    def test_step(var: str) -> str:
        return var

    response = client.get("/bob/bob")
    assert response.status_code == 200
    assert response.text == '"bob1bob"'


def test_endpoint_recovery(dbos_fastapi: Tuple[DBOS, FastAPI]) -> None:
    dbos, app = dbos_fastapi
    client = TestClient(app)

    wfuuid = str(uuid.uuid4())

    @DBOS.workflow()
    def test_workflow(var1: str) -> tuple[str, str]:
        assert DBOS.request is not None
        return var1, DBOS.workflow_id

    @app.get("/{var1}/{var2}")
    def test_endpoint(var1: str, var2: str) -> dict[str, str]:
        assert (
            DBOS.request is not None
            and DBOS.request.headers["dbos-idempotency-key"] == wfuuid
        )
        res1, id1 = test_workflow(var1)
        res2, id2 = test_workflow(var2)
        return {"res1": res1, "res2": res2, "id1": id1, "id2": id2}

    response = client.get("/a/b", headers={"dbos-idempotency-key": wfuuid})
    assert response.status_code == 200
    assert response.json().get("res1") == "a"
    assert response.json().get("res2") == "b"
    assert response.json().get("id1") == wfuuid
    assert response.json().get("id2") != wfuuid

    dbos._sys_db.wait_for_buffer_flush()
    # Change the workflow status to pending
    dbos._sys_db.update_workflow_status(
        {
            "workflow_uuid": wfuuid,
            "status": "PENDING",
            "name": test_workflow.__qualname__,
            "class_name": None,
            "config_name": None,
            "output": None,
            "error": None,
            "executor_id": None,
            "app_id": None,
            "app_version": None,
            "request": None,
            "recovery_attempts": None,
            "authenticated_user": None,
            "authenticated_roles": None,
            "assumed_role": None,
            "queue_name": None,
            "created_at": None,
            "updated_at": None,
        }
    )

    # Recovery should execute the workflow again but skip the transaction
    workflow_handles = DBOS.recover_pending_workflows()
    assert len(workflow_handles) == 1
    assert workflow_handles[0].get_result() == ("a", wfuuid)


@pytest.mark.asyncio
async def test_custom_lifespan(
    config: ConfigFile, cleanup_test_databases: None
) -> None:
    resource = None
    port = 8000

    @asynccontextmanager
    async def lifespan(app: FastAPI) -> Any:
        nonlocal resource
        resource = 1
        yield
        resource = None

    app = FastAPI(lifespan=lifespan)

    DBOS.destroy()
    DBOS(fastapi=app, config=config)

    @app.get("/")
    @DBOS.workflow()
    async def resource_workflow() -> Any:
        return {"resource": resource}

    uvicorn_config = uvicorn.Config(
        app=app, host="127.0.0.1", port=port, log_level="error"
    )
    server = uvicorn.Server(config=uvicorn_config)

    # Run server in background task
    server_task = asyncio.create_task(server.serve())
    await asyncio.sleep(0.2)  # Give server time to start

    async with httpx.AsyncClient() as client:
        r = await client.get(f"http://127.0.0.1:{port}")
        assert r.json()["resource"] == 1

    server.should_exit = True
    await server_task
    assert resource is None


def test_stacked_decorators_wf(dbos_fastapi: Tuple[DBOS, FastAPI]) -> None:
    dbos, app = dbos_fastapi
    client = TestClient(app)

    @app.get("/endpoint/{var1}/{var2}")
    @DBOS.workflow()
    async def test_endpoint(var1: str, var2: str) -> str:
        return f"{var1}, {var2}!"

    response = client.get("/endpoint/plums/deify")
    assert response.status_code == 200
    assert response.text == '"plums, deify!"'


def test_stacked_decorators_step(dbos_fastapi: Tuple[DBOS, FastAPI]) -> None:
    dbos, app = dbos_fastapi
    client = TestClient(app)

    @app.get("/endpoint/{var1}/{var2}")
    @DBOS.step()
    async def test_endpoint(var1: str, var2: str) -> str:
        return f"{var1}, {var2}!"

    response = client.get("/endpoint/plums/deify")
    assert response.status_code == 200
    assert response.text == '"plums, deify!"'



"""workflow_queue

Revision ID: d76646551a6c
Revises: d76646551a6b
Create Date: 2024-09-27 12:00:00.0

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "d76646551a6c"
down_revision: Union[str, None] = "d76646551a6b"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.rename_table("job_queue", "workflow_queue", schema="dbos")
    op.execute("CREATE VIEW dbos.job_queue AS SELECT * FROM dbos.workflow_queue;")


def downgrade() -> None:
    op.execute("DROP VIEW dbos.job_queue;")
    op.rename_table("workflow_queue", "job_queue", schema="dbos")



import uuid
from typing import Any
from urllib.parse import urlparse

from flask import Flask, request
from werkzeug.wrappers import Request as WRequest

from ._context import (
    EnterDBOSHandler,
    OperationType,
    SetWorkflowID,
    TracedAttributes,
    assert_current_dbos_context,
)
from ._request import Address, Request, request_id_header


class FlaskMiddleware:
    def __init__(self, app: Any) -> None:
        self.app = app

    def __call__(self, environ: Any, start_response: Any) -> Any:
        request = WRequest(environ)
        attributes: TracedAttributes = {
            "name": urlparse(request.url).path,
            "requestID": _get_or_generate_request_id(request),
            "requestIP": (
                request.remote_addr if request.remote_addr is not None else None
            ),
            "requestURL": request.url,
            "requestMethod": request.method,
            "operationType": OperationType.HANDLER.value,
        }
        with EnterDBOSHandler(attributes):
            ctx = assert_current_dbos_context()
            ctx.request = _make_request(request)
            workflow_id = request.headers.get("dbos-idempotency-key")
            if workflow_id is not None:
                # Set the workflow ID for the handler
                with SetWorkflowID(workflow_id):
                    response = self.app(environ, start_response)
            else:
                response = self.app(environ, start_response)
        return response


def _get_or_generate_request_id(request: WRequest) -> str:
    request_id = request.headers.get(request_id_header, None)
    if request_id is not None:
        return request_id
    else:
        return str(uuid.uuid4())


def _make_request(request: WRequest) -> Request:
    parsed_url = urlparse(request.url)
    base_url = f"{parsed_url.scheme}://{parsed_url.netloc}"

    client = None
    if request.remote_addr:
        hostname = request.remote_addr
        port = request.environ.get("REMOTE_PORT")
        if port:
            client = Address(hostname=hostname, port=int(port))
        else:
            # If port is not available, use 0 as a placeholder
            client = Address(hostname=hostname, port=0)

    return Request(
        headers=dict(request.headers),
        path_params={},
        query_params=dict(request.args),
        url=request.url,
        base_url=base_url,
        client=client,
        cookies=dict(request.cookies),
        method=request.method,
    )


def setup_flask_middleware(app: Flask) -> None:
    app.wsgi_app = FlaskMiddleware(app.wsgi_app)  # type: ignore



import os
import time
from dataclasses import dataclass
from typing import Any, Dict, Optional

import jwt
import requests
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.x509 import load_pem_x509_certificate
from rich import print

from .._logger import dbos_logger

# Constants
DBOS_CLOUD_HOST = os.getenv("DBOS_DOMAIN", "cloud.dbos.dev")
PRODUCTION_ENVIRONMENT = DBOS_CLOUD_HOST == "cloud.dbos.dev"
AUTH0_DOMAIN = "login.dbos.dev" if PRODUCTION_ENVIRONMENT else "dbos-inc.us.auth0.com"
DBOS_CLIENT_ID = (
    "6p7Sjxf13cyLMkdwn14MxlH7JdhILled"
    if PRODUCTION_ENVIRONMENT
    else "G38fLmVErczEo9ioCFjVIHea6yd0qMZu"
)
DBOS_CLOUD_IDENTIFIER = "dbos-cloud-api"


@dataclass
class DeviceCodeResponse:
    device_code: str
    user_code: str
    verification_uri: str
    verification_uri_complete: str
    expires_in: int
    interval: int

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "DeviceCodeResponse":
        return cls(
            device_code=data["device_code"],
            user_code=data["user_code"],
            verification_uri=data["verification_uri"],
            verification_uri_complete=data["verification_uri_complete"],
            expires_in=data["expires_in"],
            interval=data["interval"],
        )


@dataclass
class TokenResponse:
    access_token: str
    token_type: str
    expires_in: int
    refresh_token: Optional[str] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "TokenResponse":
        return cls(
            access_token=data["access_token"],
            token_type=data["token_type"],
            expires_in=data["expires_in"],
            refresh_token=data.get("refresh_token"),
        )


@dataclass
class AuthenticationResponse:
    token: str
    refresh_token: Optional[str] = None


class JWKSClient:
    def __init__(self, jwks_uri: str):
        self.jwks_uri = jwks_uri

    def get_signing_key(self, kid: str) -> RSAPublicKey:
        response = requests.get(self.jwks_uri)
        jwks = response.json()
        for key in jwks["keys"]:
            if key["kid"] == kid:
                cert_text = f"-----BEGIN CERTIFICATE-----\n{key['x5c'][0]}\n-----END CERTIFICATE-----"
                cert = load_pem_x509_certificate(cert_text.encode())
                return cert.public_key()  # type: ignore
        raise Exception(f"Unable to find signing key with kid: {kid}")


def verify_token(token: str) -> None:
    header = jwt.get_unverified_header(token)

    if not header.get("kid"):
        raise ValueError("Invalid token: No 'kid' in header")

    client = JWKSClient(f"https://{AUTH0_DOMAIN}/.well-known/jwks.json")
    signing_key = client.get_signing_key(header["kid"])
    jwt.decode(
        token,
        signing_key,
        algorithms=["RS256"],
        audience=DBOS_CLOUD_IDENTIFIER,
        options={
            "verify_iat": False,
            "clock_tolerance": 60,
        },
    )


def authenticate(get_refresh_token: bool = False) -> Optional[AuthenticationResponse]:
    print(
        "[bold blue]Please authenticate with DBOS Cloud to access a Postgres database[/bold blue]"
    )

    # Get device code
    device_code_data = {
        "client_id": DBOS_CLIENT_ID,
        "scope": "offline_access" if get_refresh_token else "sub",
        "audience": DBOS_CLOUD_IDENTIFIER,
    }

    try:
        response = requests.post(
            f"https://{AUTH0_DOMAIN}/oauth/device/code",
            data=device_code_data,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        device_code_response = DeviceCodeResponse.from_dict(response.json())
    except Exception as e:
        dbos_logger.error(f"Failed to log in: {str(e)}")
        return None

    login_url = device_code_response.verification_uri_complete
    print(f"[bold blue]Login URL:[/bold blue] {login_url}")

    # Poll for token
    token_data = {
        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        "device_code": device_code_response.device_code,
        "client_id": DBOS_CLIENT_ID,
    }

    elapsed_time_sec = 0
    token_response = None

    while elapsed_time_sec < device_code_response.expires_in:
        try:
            time.sleep(device_code_response.interval)
            elapsed_time_sec += device_code_response.interval

            response = requests.post(
                f"https://{AUTH0_DOMAIN}/oauth/token",
                data=token_data,
                headers={"content-type": "application/x-www-form-urlencoded"},
            )
            if response.status_code == 200:
                token_response = TokenResponse.from_dict(response.json())
                break
        except Exception:
            dbos_logger.info("Waiting for login...")

    if not token_response:
        return None

    verify_token(token_response.access_token)
    return AuthenticationResponse(
        token=token_response.access_token, refresh_token=token_response.refresh_token
    )



import os
from typing import TYPE_CHECKING, Optional

from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, ConsoleSpanExporter
from opentelemetry.trace import Span

from dbos._utils import GlobalParams

from ._dbos_config import ConfigFile

if TYPE_CHECKING:
    from ._context import TracedAttributes


class DBOSTracer:

    def __init__(self) -> None:
        self.app_id = os.environ.get("DBOS__APPID", None)
        self.provider: Optional[TracerProvider] = None

    def config(self, config: ConfigFile) -> None:
        if not isinstance(trace.get_tracer_provider(), TracerProvider):
            provider = TracerProvider()
            if os.environ.get("DBOS__CONSOLE_TRACES", None) is not None:
                processor = BatchSpanProcessor(ConsoleSpanExporter())
                provider.add_span_processor(processor)
            otlp_traces_endpoint = (
                config.get("telemetry", {}).get("OTLPExporter", {}).get("tracesEndpoint")  # type: ignore
            )
            if otlp_traces_endpoint:
                processor = BatchSpanProcessor(
                    OTLPSpanExporter(endpoint=otlp_traces_endpoint)
                )
                provider.add_span_processor(processor)
            trace.set_tracer_provider(provider)

    def set_provider(self, provider: Optional[TracerProvider]) -> None:
        self.provider = provider

    def start_span(
        self, attributes: "TracedAttributes", parent: Optional[Span] = None
    ) -> Span:
        tracer = (
            self.provider.get_tracer("dbos-tracer")
            if self.provider is not None
            else trace.get_tracer("dbos-tracer")
        )
        context = trace.set_span_in_context(parent) if parent else None
        span: Span = tracer.start_span(name=attributes["name"], context=context)
        attributes["applicationID"] = self.app_id
        attributes["applicationVersion"] = GlobalParams.app_version
        attributes["executorID"] = GlobalParams.executor_id
        for k, v in attributes.items():
            if k != "name" and v is not None and isinstance(v, (str, bool, int, float)):
                span.set_attribute(k, v)
        return span

    def end_span(self, span: Span) -> None:
        span.end()


dbos_tracer = DBOSTracer()



from dataclasses import dataclass
from typing import Any, Mapping, NamedTuple, Optional

request_id_header = "x-request-id"


class Address(NamedTuple):
    hostname: str
    port: int


@dataclass
class Request:
    """
    Serializable HTTP Request object.

    Attributes:
        base_url(str): Base of URL requested, as in application code
        client(Optional[Address]): HTTP Client
        cookies(Mapping[str, str]): HTTP Cookies
        headers(Mapping[str, str]): HTTP headers
        method(str): HTTP verb
        path_params(Mapping[str, Any]): Parameters extracted from URL path sections
        query_params(Mapping[str, str]): URL query string parameters
        url(str): Full URL accessed
    """

    headers: Mapping[str, str]
    path_params: Mapping[str, Any]
    query_params: Mapping[str, str]
    url: str
    base_url: str
    client: Optional[Address]
    cookies: Mapping[str, str]
    method: str



from sqlalchemy import (
    BigInteger,
    Boolean,
    Column,
    ForeignKey,
    Index,
    Integer,
    MetaData,
    PrimaryKeyConstraint,
    String,
    Table,
    Text,
    text,
)


class SystemSchema:
    ### System table schema
    metadata_obj = MetaData(schema="dbos")
    sysdb_suffix = "_dbos_sys"

    workflow_status = Table(
        "workflow_status",
        metadata_obj,
        Column("workflow_uuid", Text, primary_key=True),
        Column("status", Text, nullable=True),
        Column("name", Text, nullable=True),
        Column("authenticated_user", Text, nullable=True),
        Column("assumed_role", Text, nullable=True),
        Column("authenticated_roles", Text, nullable=True),
        Column("request", Text, nullable=True),
        Column("output", Text, nullable=True),
        Column("error", Text, nullable=True),
        Column("executor_id", Text, nullable=True),
        Column(
            "created_at",
            BigInteger,
            nullable=False,
            server_default=text("(EXTRACT(epoch FROM now()) * 1000::numeric)::bigint"),
        ),
        Column(
            "updated_at",
            BigInteger,
            nullable=False,
            server_default=text("(EXTRACT(epoch FROM now()) * 1000::numeric)::bigint"),
        ),
        Column("application_version", Text, nullable=True),
        Column("application_id", Text, nullable=True),
        Column("class_name", String(255), nullable=True, server_default=text("NULL")),
        Column("config_name", String(255), nullable=True, server_default=text("NULL")),
        Column(
            "recovery_attempts",
            BigInteger,
            nullable=True,
            server_default=text("'0'::bigint"),
        ),
        Column("queue_name", Text),
        Index("workflow_status_created_at_index", "created_at"),
        Index("workflow_status_executor_id_index", "executor_id"),
    )

    operation_outputs = Table(
        "operation_outputs",
        metadata_obj,
        Column(
            "workflow_uuid",
            Text,
            ForeignKey(
                "workflow_status.workflow_uuid", onupdate="CASCADE", ondelete="CASCADE"
            ),
            nullable=False,
        ),
        Column("function_id", Integer, nullable=False),
        Column("output", Text, nullable=True),
        Column("error", Text, nullable=True),
        PrimaryKeyConstraint("workflow_uuid", "function_id"),
    )

    workflow_inputs = Table(
        "workflow_inputs",
        metadata_obj,
        Column(
            "workflow_uuid",
            Text,
            ForeignKey(
                "workflow_status.workflow_uuid", onupdate="CASCADE", ondelete="CASCADE"
            ),
            primary_key=True,
            nullable=False,
        ),
        Column("inputs", Text, nullable=False),
    )

    notifications = Table(
        "notifications",
        metadata_obj,
        Column(
            "destination_uuid",
            Text,
            ForeignKey(
                "workflow_status.workflow_uuid", onupdate="CASCADE", ondelete="CASCADE"
            ),
            nullable=False,
        ),
        Column("topic", Text, nullable=True),
        Column("message", Text, nullable=False),
        Column(
            "created_at_epoch_ms",
            BigInteger,
            nullable=False,
            server_default=text("(EXTRACT(epoch FROM now()) * 1000::numeric)::bigint"),
        ),
        Column(
            "message_uuid",
            Text,
            nullable=False,
            server_default=text("uuid_generate_v4()"),
        ),
        Index("idx_workflow_topic", "destination_uuid", "topic"),
    )

    workflow_events = Table(
        "workflow_events",
        metadata_obj,
        Column(
            "workflow_uuid",
            Text,
            ForeignKey(
                "workflow_status.workflow_uuid", onupdate="CASCADE", ondelete="CASCADE"
            ),
            nullable=False,
        ),
        Column("key", Text, nullable=False),
        Column("value", Text, nullable=False),
        PrimaryKeyConstraint("workflow_uuid", "key"),
    )

    scheduler_state = Table(
        "scheduler_state",
        metadata_obj,
        Column("workflow_fn_name", Text, primary_key=True, nullable=False),
        Column("last_run_time", BigInteger, nullable=False),
    )

    workflow_queue = Table(
        "workflow_queue",
        metadata_obj,
        Column(
            "workflow_uuid",
            Text,
            ForeignKey(
                "workflow_status.workflow_uuid", onupdate="CASCADE", ondelete="CASCADE"
            ),
            nullable=False,
            primary_key=True,
        ),
        # Column("executor_id", Text), # This column is deprecated. Do *not* use it.
        Column("queue_name", Text, nullable=False),
        Column(
            "created_at_epoch_ms",
            BigInteger,
            nullable=False,
            server_default=text("(EXTRACT(epoch FROM now()) * 1000::numeric)::bigint"),
        ),
        Column(
            "started_at_epoch_ms",
            BigInteger(),
        ),
        Column(
            "completed_at_epoch_ms",
            BigInteger(),
        ),
    )



import json
from dataclasses import asdict, dataclass
from enum import Enum
from typing import List, Optional, Type, TypedDict, TypeVar

from dbos._workflow_commands import WorkflowInformation


class MessageType(str, Enum):
    EXECUTOR_INFO = "executor_info"
    RECOVERY = "recovery"
    CANCEL = "cancel"
    LIST_WORKFLOWS = "list_workflows"
    LIST_QUEUED_WORKFLOWS = "list_queued_workflows"
    RESUME = "resume"
    RESTART = "restart"
    GET_WORKFLOW = "get_workflow"


T = TypeVar("T", bound="BaseMessage")


@dataclass
class BaseMessage:
    type: MessageType
    request_id: str

    @classmethod
    def from_json(cls: Type[T], json_str: str) -> T:
        """
        Safely load a JSON into a dataclass, loading only the
        attributes specified in the dataclass.
        """
        data = json.loads(json_str)
        all_annotations = {}
        for base_cls in cls.__mro__:
            if hasattr(base_cls, "__annotations__"):
                all_annotations.update(base_cls.__annotations__)
        kwargs = {k: v for k, v in data.items() if k in all_annotations}
        return cls(**kwargs)

    def to_json(self) -> str:
        dict_data = asdict(self)
        return json.dumps(dict_data)


@dataclass
class ExecutorInfoRequest(BaseMessage):
    pass


@dataclass
class ExecutorInfoResponse(BaseMessage):
    executor_id: str
    application_version: str


@dataclass
class RecoveryRequest(BaseMessage):
    executor_ids: List[str]


@dataclass
class RecoveryResponse(BaseMessage):
    success: bool


@dataclass
class CancelRequest(BaseMessage):
    workflow_id: str


@dataclass
class CancelResponse(BaseMessage):
    success: bool


@dataclass
class ResumeRequest(BaseMessage):
    workflow_id: str


@dataclass
class ResumeResponse(BaseMessage):
    success: bool


@dataclass
class RestartRequest(BaseMessage):
    workflow_id: str


@dataclass
class RestartResponse(BaseMessage):
    success: bool


class ListWorkflowsBody(TypedDict):
    workflow_uuids: List[str]
    workflow_name: Optional[str]
    authenticated_user: Optional[str]
    start_time: Optional[str]
    end_time: Optional[str]
    status: Optional[str]
    application_version: Optional[str]
    limit: Optional[int]
    offset: Optional[int]
    sort_desc: bool


@dataclass
class WorkflowsOutput:
    WorkflowUUID: str
    Status: Optional[str]
    WorkflowName: Optional[str]
    WorkflowClassName: Optional[str]
    WorkflowConfigName: Optional[str]
    AuthenticatedUser: Optional[str]
    AssumedRole: Optional[str]
    AuthenticatedRoles: Optional[str]
    Input: Optional[str]
    Output: Optional[str]
    Request: Optional[str]
    Error: Optional[str]
    CreatedAt: Optional[str]
    UpdatedAt: Optional[str]
    QueueName: Optional[str]
    ApplicationVersion: Optional[str]

    @classmethod
    def from_workflow_information(cls, info: WorkflowInformation) -> "WorkflowsOutput":
        # Convert fields to strings as needed
        created_at_str = str(info.created_at) if info.created_at is not None else None
        updated_at_str = str(info.updated_at) if info.updated_at is not None else None
        inputs_str = str(info.input) if info.input is not None else None
        outputs_str = str(info.output) if info.output is not None else None
        request_str = str(info.request) if info.request is not None else None

        return cls(
            WorkflowUUID=info.workflow_id,
            Status=info.status,
            WorkflowName=info.workflow_name,
            WorkflowClassName=info.workflow_class_name,
            WorkflowConfigName=info.workflow_config_name,
            AuthenticatedUser=info.authenticated_user,
            AssumedRole=info.assumed_role,
            AuthenticatedRoles=info.authenticated_roles,
            Input=inputs_str,
            Output=outputs_str,
            Request=request_str,
            Error=info.error,
            CreatedAt=created_at_str,
            UpdatedAt=updated_at_str,
            QueueName=info.queue_name,
            ApplicationVersion=info.app_version,
        )


@dataclass
class ListWorkflowsRequest(BaseMessage):
    body: ListWorkflowsBody


@dataclass
class ListWorkflowsResponse(BaseMessage):
    output: List[WorkflowsOutput]


class ListQueuedWorkflowsBody(TypedDict):
    workflow_name: Optional[str]
    start_time: Optional[str]
    end_time: Optional[str]
    status: Optional[str]
    queue_name: Optional[str]
    limit: Optional[int]
    offset: Optional[int]
    sort_desc: bool


@dataclass
class ListQueuedWorkflowsRequest(BaseMessage):
    body: ListQueuedWorkflowsBody


@dataclass
class ListQueuedWorkflowsResponse(BaseMessage):
    output: List[WorkflowsOutput]


@dataclass
class GetWorkflowRequest(BaseMessage):
    workflow_id: str


@dataclass
class GetWorkflowResponse(BaseMessage):
    output: Optional[WorkflowsOutput]



import threading
import time
import uuid
from concurrent.futures import Future, ThreadPoolExecutor
from typing import Tuple

from sqlalchemy import text

# Public API
from dbos import DBOS, SetWorkflowID


def test_concurrent_workflows(dbos: DBOS) -> None:
    @DBOS.workflow()
    def test_workflow() -> str:
        time.sleep(1)
        return DBOS.workflow_id

    def test_thread(id: str) -> str:
        with SetWorkflowID(id):
            return test_workflow()

    num_threads = 10
    with ThreadPoolExecutor(max_workers=num_threads) as executor:
        futures: list[Tuple[str, Future[str]]] = []
        for _ in range(num_threads):
            id = str(uuid.uuid4())
            futures.append((id, executor.submit(test_thread, id)))
        for id, future in futures:
            assert id == future.result()


def test_concurrent_conflict_uuid(dbos: DBOS) -> None:
    condition = threading.Condition()
    step_count = 0
    txn_count = 0

    @DBOS.step()
    def test_step() -> str:
        nonlocal step_count
        condition.acquire()
        step_count += 1
        if step_count % 2 == 1:
            # Wait for the other one to notify
            condition.wait()
        else:
            # Notify the other one
            condition.notify()
        condition.release()

        return DBOS.workflow_id

    @DBOS.workflow()
    def test_workflow() -> str:
        res = test_step()
        return res

    def test_comm_thread(id: str) -> str:
        with SetWorkflowID(id):
            return test_step()

    # Need to set isolation level to a lower one, otherwise it gets serialization error instead (we already handle it correctly by automatic retries).
    @DBOS.transaction(isolation_level="REPEATABLE READ")
    def test_transaction() -> str:
        DBOS.sql_session.execute(text("SELECT 1")).fetchall()
        nonlocal txn_count
        condition.acquire()
        txn_count += 1
        if txn_count % 2 == 1:
            # Wait for the other one to notify
            condition.wait()
        else:
            # Notify the other one
            condition.notify()
        condition.release()

        return DBOS.workflow_id

    def test_txn_thread(id: str) -> str:
        with SetWorkflowID(id):
            return test_transaction()

    wfuuid = str(uuid.uuid4())
    with SetWorkflowID(wfuuid):
        wf_handle1 = dbos.start_workflow(test_workflow)

    with SetWorkflowID(wfuuid):
        wf_handle2 = dbos.start_workflow(test_workflow)

    # These two workflows should run concurrently, but both should succeed.
    assert wf_handle1.get_result() == wfuuid
    assert wf_handle2.get_result() == wfuuid

    # Make sure temp workflows can handle conflicts as well.
    wfuuid = str(uuid.uuid4())
    with ThreadPoolExecutor(max_workers=2) as executor:
        future1 = executor.submit(test_comm_thread, wfuuid)
        future2 = executor.submit(test_comm_thread, wfuuid)

    assert future1.result() == wfuuid
    assert future2.result() == wfuuid

    # Make sure temp transactions can handle conflicts as well.
    wfuuid = str(uuid.uuid4())
    with ThreadPoolExecutor(max_workers=2) as executor:
        future1 = executor.submit(test_txn_thread, wfuuid)
        future2 = executor.submit(test_txn_thread, wfuuid)

    assert future1.result() == wfuuid
    assert future2.result() == wfuuid



"""
Add triggers.

Revision ID: a3b18ad34abe
Revises: 5c361fc04708
Create Date: 2024-07-21 13:07:09.814989
"""

from typing import Sequence, Union

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "a3b18ad34abe"
down_revision: Union[str, None] = "5c361fc04708"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    op.execute(
        """
      CREATE OR REPLACE FUNCTION dbos.notifications_function() RETURNS TRIGGER AS $$
      DECLARE
          payload text := NEW.destination_uuid || '::' || NEW.topic;
      BEGIN
          PERFORM pg_notify('dbos_notifications_channel', payload);
          RETURN NEW;
      END;
      $$ LANGUAGE plpgsql;
               """
    )
    op.execute(
        """
      CREATE TRIGGER dbos_notifications_trigger
      AFTER INSERT ON dbos.notifications
      FOR EACH ROW EXECUTE FUNCTION dbos.notifications_function();
               """
    )
    op.execute(
        """
      CREATE OR REPLACE FUNCTION dbos.workflow_events_function() RETURNS TRIGGER AS $$
      DECLARE
          payload text := NEW.workflow_uuid || '::' || NEW.key;
      BEGIN
          PERFORM pg_notify('dbos_workflow_events_channel', payload);
          RETURN NEW;
      END;
      $$ LANGUAGE plpgsql;
               """
    )
    op.execute(
        """
    CREATE TRIGGER dbos_workflow_events_trigger
      AFTER INSERT ON dbos.workflow_events
      FOR EACH ROW EXECUTE FUNCTION dbos.workflow_events_function();
               """
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    op.execute(
        "DROP TRIGGER IF EXISTS dbos_notifications_trigger ON dbos.notifications;"
    )
    op.execute("DROP FUNCTION IF EXISTS dbos.notifications_function;")
    op.execute(
        "DROP TRIGGER IF EXISTS dbos_workflow_events_trigger ON dbos.workflow_events;"
    )
    op.execute("DROP FUNCTION IF EXISTS dbos.workflow_events_function;")
    # ### end Alembic commands ###



import re
import runpy
import sys
from typing import Union

from dbos import DBOS


class PythonModule:
    def __init__(self, module_name: str):
        self.module_name = module_name


def debug_workflow(workflow_id: str, entrypoint: Union[str, PythonModule]) -> None:
    # include the current directory (represented by empty string) in the search path
    # if it not already included
    if "" not in sys.path:
        sys.path.insert(0, "")
    if isinstance(entrypoint, str):
        runpy.run_path(entrypoint)
    elif isinstance(entrypoint, PythonModule):
        runpy.run_module(entrypoint.module_name)
    else:
        raise ValueError("Invalid entrypoint type. Must be a string or PythonModule.")

    DBOS.logger.info(f"Debugging workflow {workflow_id}...")
    DBOS.launch(debug_mode=True)
    handle = DBOS.execute_workflow_id(workflow_id)
    handle.get_result()
    DBOS.logger.info("Workflow Debugging complete. Exiting process.")


def parse_start_command(command: str) -> Union[str, PythonModule]:
    match = re.match(r"fastapi\s+run\s+(\.?[\w/]+\.py)", command)
    if match:
        return match.group(1)
    match = re.match(r"python3?\s+(\.?[\w/]+\.py)", command)
    if match:
        return match.group(1)
    match = re.match(r"python3?\s+-m\s+([\w\.]+)", command)
    if match:
        return PythonModule(match.group(1))
    raise ValueError(
        "Invalid command format. Must be 'fastapi run <script>' or 'python <script>' or 'python -m <module>'"
    )
